#include "declare.h"
#include "editor.h"
#include "Argtricks.h"
#include <X11/Wc/WcCreate.h>
#include <sys/stat.h>

#include <Xm/MainW.h>
#include <Xm/CutPaste.h>

/* do not turn this off - it was for testing purposes */
#define WIDE_WORK_PIXMAP

#define IS_NORMAL(w)	(strcmp(XtName(w), "normal") == 0)
#define IS_ZOOM(w)	(strcmp(XtName(w), "zoom") == 0)
#define MAX_PTS		256
#define MAX_PIXMAP_SIZE	9999

#ifdef _NO_PROTO
void editorCreate();
#else
void editorCreate(Widget parent, String filename);
#endif

String pixFiles[] = { "draw.xpm", "line.xpm", "rect.xpm", "fillrect.xpm",
	"circle.xpm", "disc.xpm", "flood.xpm", "bitmap.xpm", "text.xpm",
	"copy.xpm" };

#ifdef _NO_PROTO
void loadPixmap(wid, filename, pixRes, bgRes)
Widget wid; String filename; String pixRes; String bgRes;
#else
void loadPixmap(Widget wid, String filename, String pixRes, String bgRes)
#endif
{
char name[256];
    strcpy(name, "/home/msm/doyle/xstuff/xpe/");
    strcat(name, filename);

    /* try /home/msm/doyle/xstuff/xpe first */
    if (XLoadExplicitWidgetPixmap(wid, RootWindow(display, screen),
	DefaultVisual(display, screen), DefaultDepth(display, screen),
	cmap, name, pixRes, bgRes))
    {
	/* else try current directory */
	if (XLoadExplicitWidgetPixmap(wid, RootWindow(display, screen),
	    DefaultVisual(display, screen), DefaultDepth(display, screen),
	    cmap, filename, pixRes, bgRes))
	{
	    printf("Cannot find needed pixmap file '%s' For Icon\n", filename);
	    exit(-1);
	}
    }
}

/* ARGSUSED */
#ifdef _NO_PROTO
static void setLabelPixmapCB(wid, unused, unused2)
Widget wid; XtPointer unused; XtPointer unused2;
#else
static void setLabelPixmapCB(Widget wid, XtPointer unused, XtPointer unused2)
#endif
{
    String name, filename;

    name = XtName(wid);
    if (strcmp(name, "primDraw") == 0)
	filename = "draw.xpm";
    else if (strcmp(name, "primLine") == 0)
	filename = "line.xpm";
    else if (strcmp(name, "primRect") == 0)
	filename = "rect.xpm";
    else if (strcmp(name, "primFillRect") == 0)
	filename = "fillrect.xpm";
    else if (strcmp(name, "primCircle") == 0)
	filename = "circle.xpm";
    else if (strcmp(name, "primFillCircle") == 0)
	filename = "disc.xpm";
    else if (strcmp(name, "primFlood") == 0)
	filename = "flood.xpm";
    else if (strcmp(name, "primBitmap") == 0)
	filename = "bitmap.xpm";
    else if (strcmp(name, "primText") == 0)
	filename = "text.xpm";
    else if (strcmp(name, "primCopy") == 0)
	filename = "copy.xpm";
    else if (strcmp(name, "primMove") == 0)
	filename = "move.xpm";
    else if (strcmp(name, "primPaste") == 0)
	filename = "paste.xpm";
    else
	return;
    loadPixmap(wid, filename, XmNlabelPixmap, XmNbackground);
}

#ifdef _NO_PROTO
EditorPtr getEditor(wid)
Widget wid;
#else
EditorPtr getEditor(Widget wid)
#endif
{
    EditorPtr ep = 0;
    DeclareArgs(1);

    while (wid && !XmIsMainWindow(wid))
	wid = XtParent(wid);
    if (!wid)
	return(NULL);
    FirstArg(XmNuserData, &ep);
    XtGetValues(wid, Args, ArgCount);
    return(ep);
}

#ifdef _NO_PROTO
void sayMessage(ep, msg)
EditorPtr ep; String msg;
#else
void sayMessage(EditorPtr ep, String msg)
#endif
{
    XmTextFieldSetString(ep->messageArea, msg);
    XmUpdateDisplay(topLevel);
}

#ifdef _NO_PROTO
static void showPixmap(wid, e)
Widget wid; XExposeEvent *e;
#else
static void showPixmap(Widget wid, XExposeEvent *e)
#endif
{
int x, y, xmin, xmax, ymin, ymax;
unsigned long fg, lastfg;
Window win = XtWindow(wid);
XRectangle rects[MAX_PTS];
int nr = 0;

    EditorPtr ep;
    ep = getEditor(wid);
    if (IS_NORMAL(wid))
    {
	XCopyArea(display, ep->pixmap, win, ep->gc, 
		0, 0, ep->pw, ep->ph, 0, 0);
	return;
    }

    lastfg = ep->workbg;
    XSetForeground(display, ep->gc, lastfg);
    if (e)
	XFillRectangle(display, win, ep->gc, e->x, e->y, e->width, e->height);
    else
	XFillRectangle(display, win, ep->gc, 0, 0, 
		1 + ep->pw * (ep->scale + 1), 1 + ep->ph * (ep->scale + 1));

    if (e)
    {
	xmin = e->x / (ep->scale + 1);
	ymin = e->y / (ep->scale + 1);
	xmax = xmin + 2 + (e->width / (ep->scale + 1));
	if (xmax > ep->pw)
	    xmax = ep->pw;
	ymax = ymin + 2 + (e->height / (ep->scale + 1));
	if (ymax > ep->ph)
	    ymax = ep->ph;
    }
    else
    {
	xmin = 0;
	xmax = ep->pw;
	ymin = 0;
	ymax = ep->ph;
    }

    for (x=xmin; x<xmax; x++)
    {
	for (y=ymin; y<ymax; y++)
	{
	    fg = XGetPixel(ep->image, x, y);
	    if (lastfg != fg)
	    {
		if (nr)
		{
		    XFillRectangles(display, win, ep->gc, rects, nr);
		    nr = 0;
		}
		XSetForeground(display, ep->gc, lastfg = fg);
	    }
	    rects[nr].x = 1 + x * (ep->scale + 1);
	    rects[nr].y = 1 + y * (ep->scale + 1);
	    rects[nr].width = rects[nr].height = ep->scale;
	    if (++nr == MAX_PTS)
	    {
		XFillRectangles(display, win, ep->gc, rects, nr);
		    nr = 0;
	    }
	}
    }
    if (nr)
	XFillRectangles(display, win, ep->gc, rects, nr);
}

/* ARGSUSED */
#ifdef _NO_PROTO
static void redrawImageCB(wid, unused, cb)
Widget wid; String unused; XmAnyCallbackStruct *cb;
#else
static void redrawImageCB(Widget wid, String unused, XmAnyCallbackStruct *cb)
#endif
{
    if (!XtIsRealized(wid))
	return;
    if (cb->event->type != Expose)
	return;
    showPixmap(wid, &(cb->event->xexpose));
}

#ifdef _NO_PROTO
static void addEventHandler(wid, ep)
Widget wid; EditorPtr ep;
#else
static void addEventHandler(Widget wid, EditorPtr ep)
#endif
{
extern void configInputEH();
extern void inputImageEH();

    if (IS_ZOOM(wid))
    {

	XtAddEventHandler(ep->clipWidget, StructureNotifyMask,
	    False, (XtEventHandler)configInputEH, (XtPointer)ep);

	XtAddEventHandler(wid, StructureNotifyMask | VisibilityChangeMask,
	    False, (XtEventHandler)configInputEH, (XtPointer)ep);
	XtAddEventHandler(wid,
	    ButtonPressMask | ButtonReleaseMask | ButtonMotionMask,
	    False, (XtEventHandler)inputImageEH, (XtPointer)ep);
    }
    else if (IS_NORMAL(wid))
    {
	XtAddEventHandler(wid,
	    ButtonPressMask | ButtonReleaseMask | ButtonMotionMask,
	    False, (XtEventHandler)inputImageEH, (XtPointer)ep);
    }
}


/*
 * write file to disk
 */
#ifdef _NO_PROTO
void savePixFile(ep, newFileName)
EditorPtr ep; String newFileName;
#else
void savePixFile(EditorPtr ep, String newFileName)
#endif
{
    DeclareArgs(1);
    int status;
    char *ct[256][6];		/* I need 256 of array of 6 string ptrs */
    char **ctPtrs[256];		/* XPM requires array of ptrs to ptrs */
    Pixel ctPixels[256];
#ifdef _NO_PROTO
    String s, sn, getSymbolicName(), getRGBText();
#else
    String s, sn, getSymbolicName(int), getRGBText(int, int, int);
#endif
    int red, green, blue, changable;
    int i, numct;
    char ***saveCT;
    int saveNC;
    unsigned long *savePixels;

    /* save original info */
    saveCT = ep->info.colorTable;
    saveNC = ep->info.ncolors;
    savePixels = ep->info.pixels;

    for (numct=i=0; i<256; i++)
    {
        sn = getSymbolicName(i);
	/* if (sn) */
	{
            getRGB(i, &red, &green, &blue, &changable);
	    red >>= 8;
	    green >>= 8;
	    blue >>= 8;
            s = getRGBText(red, green, blue);
            ct[numct][0] = ct[numct][5] = XtNewString(s);
            ct[numct][1] = sn;
            ct[numct][2] = ct[numct][3] = ct[numct][4] = NULL;
            ctPixels[numct] = i;
	    ctPtrs[numct] = ct[numct];
            numct++;
	}
    }

    ep->info.colorTable = (char ***)ctPtrs;
    ep->info.ncolors = numct;
    ep->info.pixels = ctPixels;
    ep->info.width = ep->pw;
    ep->info.height = ep->ph;
    ep->info.npixels = 256;

    if (newFileName)
    {
	XtFree(ep->pixfile);
	ep->pixfile = XtNewString(newFileName);
    }

    status = XpmWritePixmapFile(display, ep->pixfile, 
	ep->pixmap, NULL, &ep->info);

    i = numct;
    while (i > 0)
    {
	i--;
	XtFree(ct[i][0]);
    }

    /* restore original info */
    ep->info.colorTable = saveCT;
    ep->info.ncolors = saveNC;
    ep->info.pixels = savePixels;


    if (status != PixmapSuccess)
    {
	char buf[1024];
	sprintf(buf, "Cannot Save Pixmap File\n'%s'", ep->pixfile);
	popupMessage(XPE_Error, buf, NULL, NULL, NULL);
    }
    FirstArg(XmNtitle, ep->pixfile);
    XtSetValues(ep->shell, Args, ArgCount);
    sayMessage(ep, "Pixmap saved");
}

/* ARGSUSED */
#ifdef _NO_PROTO
static void fontCB(wid, unused, cb)
Widget wid; String unused; XmAnyCallbackStruct *cb;
#else
static void fontCB(Widget wid, String unused, XmAnyCallbackStruct *cb)
#endif
{
    extern void selectFont();

    selectFont();
}

/* ARGSUSED */
#ifdef _NO_PROTO
static void saveCB(wid, unused, cb)
Widget wid; String unused; XmAnyCallbackStruct *cb;
#else
static void saveCB(Widget wid, String unused, XmAnyCallbackStruct *cb)
#endif
{
    EditorPtr ep;

    ep = getEditor(wid);
    if (!ep->pixfile)
	askFile(savePixFile, "Save Pixmap", ep, "*.xpm");
    else
	savePixFile(ep, NULL);
	
}

/* ARGSUSED */
#ifdef _NO_PROTO
static void saveAsCB(wid, unused, cb)
Widget wid; String unused; XmAnyCallbackStruct *cb;
#else
static void saveAsCB(Widget wid, String unused, XmAnyCallbackStruct *cb)
#endif
{
    EditorPtr ep;

    ep = getEditor(wid);
    askFile(savePixFile, "Save As Pixmap", ep, "*.xpm");
}


/* ARGSUSED */
#ifdef _NO_PROTO
static void exitCB(wid, unused, cb)
Widget wid; String unused; XmAnyCallbackStruct *cb;
#else
static void exitCB(Widget wid, String unused, XmAnyCallbackStruct *cb)
#endif
{
    exit(0);
}

/* ARGSUSED */
#ifdef _NO_PROTO
static void newCB(wid, unused, cb)
Widget wid; String unused; XmAnyCallbackStruct *cb;
#else
static void newCB(Widget wid, String unused, XmAnyCallbackStruct *cb)
#endif
{
    editorCreate(topLevel, NULL);
}

/* ARGSUSED */
#ifdef _NO_PROTO
static void openCB(wid, unused, cb)
Widget wid; String unused; XmAnyCallbackStruct *cb;
#else
static void openCB(Widget wid, String unused, XmAnyCallbackStruct *cb)
#endif
{
    askFile(editorCreate, "Open Pixmap", topLevel, "*.xpm");
}

/* ARGSUSED */
#ifdef _NO_PROTO
static void closeCB(wid, unused, cb)
Widget wid; String unused; XmAnyCallbackStruct *cb;
#else
static void closeCB(Widget wid, String unused, XmAnyCallbackStruct *cb)
#endif
{
    EditorPtr ep;

    ep = getEditor(wid);
    editorDestroy(ep);
}

#ifdef _NO_PROTO
void applyEditInfo(ep, w, h, s)
EditorPtr ep; int w; int h; int s;
#else
void applyEditInfo(EditorPtr ep, int w, int h, int s)
#endif
{
int newSize = False;
Pixmap newpix;
DeclareArgs(2);

    if (w < 1 || w > MAX_PIXMAP_SIZE || h < 1 || h > MAX_PIXMAP_SIZE)
    {
	popupMessage(XPE_Warning, 
	    "Pixmap Size is Invalid!\nChanges Ignored",
	    NULL, NULL, NULL);
	return;
    }
    if (s != ep->scale)
    {
	Widget hbar, vbar;
	int inc;

	newSize = True;
	ep->scale = s;
	inc = ep->scale + 1;

	FirstArg(XmNhorizontalScrollBar, &hbar);
	NextArg(XmNverticalScrollBar, &vbar);
	XtGetValues(ep->workScrollWin, Args, ArgCount);

	FirstArg(XmNincrement, inc);
	XtSetValues(hbar, Args, ArgCount);
	FirstArg(XmNincrement, inc);
	XtSetValues(vbar, Args, ArgCount);
    }
    if (w != ep->pw || h != ep->ph)
    {
	newSize = True;

	newpix = XCreatePixmap(display, XtWindow(ep->workarea), w, h, 
	    DefaultDepth(display, screen));
	XFreePixmap(display, ep->workpixmap);
	ep->workpixmap = XCreatePixmap(display, XtWindow(ep->workarea), w, h, 
#ifdef WIDE_WORK_PIXMAP
	    DefaultDepth(display, screen));
#else
	    1);
#endif

	XSetForeground(display, ep->gc, 0);
	XFillRectangle(display, newpix, ep->gc, 0, 0, w, h);
	XCopyArea(display, ep->pixmap, newpix, ep->gc, 0, 0, 
		ep->pw, ep->ph, 0, 0);
	XFreePixmap(display, ep->pixmap);
	ep->pixmap = newpix;

	XDestroyImage(ep->image);
	XDestroyImage(ep->workimage);
	XDestroyImage(ep->undoimage);
	ep->image = XGetImage(display, ep->pixmap, 0, 0, ep->pw=w, ep->ph=h,
			AllPlanes, ZPixmap);
	ep->workimage = XSubImage(ep->image, 0, 0, ep->pw, ep->ph);
	ep->undoimage = XSubImage(ep->image, 0, 0, ep->pw, ep->ph);
    }
    if (newSize)
    {
	FirstArg(XmNwidth, (ep->pw * (ep->scale + 1) + 1));
	NextArg(XmNheight, (ep->ph * (ep->scale + 1) + 1));
	XtSetValues(ep->workarea, Args, ArgCount);

	FirstArg(XmNwidth, ep->pw);
	NextArg(XmNheight, ep->ph);
	XtSetValues(ep->pixarea, Args, ArgCount);

	redrawWidget(ep->workarea);
	redrawWidget(ep->pixarea);
    }
}

/* ARGSUSED */
#ifdef _NO_PROTO
static void editInfoCB(wid, unused, cb)
Widget wid; String unused; XmAnyCallbackStruct *cb;
#else
static void editInfoCB(Widget wid, String unused, XmAnyCallbackStruct *cb)
#endif
{
    EditorPtr ep;

    ep = getEditor(wid);
    askSize(applyEditInfo, "Size Info", ep, ep->pw, ep->ph, ep->scale);
}


/* ARGSUSED */
#ifdef _NO_PROTO
static void fillToBoundaryCB(wid, unused, cb)
Widget wid; String unused; XmToggleButtonCallbackStruct *cb;
#else
static void fillToBoundaryCB(Widget wid, String unused, 
	XmToggleButtonCallbackStruct *cb)
#endif
{
    EditorPtr ep;

    ep = getEditor(wid);
    ep->fillToBoundary = cb->set;
}

/*
 * create the empty work pixmap
 */
#ifdef _NO_PROTO
void createPixmap(ep)
EditorPtr ep;
#else
void createPixmap(EditorPtr ep)
#endif
{
    DeclareArgs(2);

    ep->pixmap = XCreatePixmap(display, 
	RootWindowOfScreen(XtScreenOfObject(ep->shell)), ep->pw, ep->ph, 
	DefaultDepth(display, screen));
    ep->workpixmap = XCreatePixmap(display, 
	RootWindowOfScreen(XtScreenOfObject(ep->shell)),
#ifdef WIDE_WORK_PIXMAP
	ep->pw, ep->ph, DefaultDepth(display, screen));
#else
	ep->pw, ep->ph, 1);
#endif
    XSetForeground(display, ep->gc, 0);
    XFillRectangle(display, ep->pixmap, ep->gc, 0, 0, ep->pw, ep->ph);
    ep->image = XGetImage(display, ep->pixmap, 0, 0, ep->pw, ep->ph, AllPlanes, ZPixmap);
    ep->workimage = XSubImage(ep->image, 0, 0, ep->pw, ep->ph);
    ep->undoimage = XSubImage(ep->image, 0, 0, ep->pw, ep->ph);

    FirstArg(XmNwidth, (ep->pw * (ep->scale + 1)) + 1);
    NextArg(XmNheight, (ep->ph * (ep->scale + 1)) + 1);
    XtSetValues(ep->workarea, Args, ArgCount);

    FirstArg(XmNwidth, ep->pw);
    NextArg(XmNheight, ep->ph);
    XtSetValues(ep->pixarea, Args, ArgCount);

    ep->info.ncolors = 0;
    ep->info.colorTable = NULL;
    ep->info.pixels = NULL;
    ep->info.npixels = 0;
    ep->info.hints_cmt = NULL;
    ep->info.colors_cmt = NULL;
    ep->info.pixels_cmt = NULL;
    ep->info.rgb_fname = RGB_FNAME;
    ep->info.width = ep->pw;
    ep->info.height = ep->ph;
    ep->info.x_hotspot = 0;
    ep->info.y_hotspot = 0;
    ep->info.visual = DefaultVisual(display, screen);
    ep->info.colormap = cmap;
    ep->info.depth = 8;
    ep->info.mask_pixel = 0xffff;
    ep->info.valuemask = XpmSize | XpmHotspot | XpmColorSymbols 
	| XpmVisual | XpmColormap | XpmDepth
	| XpmRgbFilename | XpmInfos | XpmReturnPixels | XpmReturnInfos;

    ep->pixfile = NULL;
}

/*
 * read a pixmap from file
 */
#ifdef _NO_PROTO
void readPixmap(ep, filename)
EditorPtr ep; String filename;
#else
void readPixmap(EditorPtr ep, String filename)
#endif
{
DeclareArgs(2);
int status;

    ep->info.visual = DefaultVisual(display, screen);
    ep->info.colormap = cmap;
    ep->info.depth = DefaultDepth(display, screen);
    ep->info.rgb_fname = RGB_FNAME;
    ep->info.valuemask = XpmVisual | XpmColormap | XpmDepth |
	XpmSize | XpmHotspot | XpmColorSymbols | XpmRgbFilename | XpmInfos
	| XpmReturnPixels | XpmReturnInfos;

    status = XpmReadPixmapFile(display, 
	RootWindowOfScreen(XtScreenOfObject(ep->shell)),
	filename, &ep->pixmap, NULL, &ep->info);

    ep->pw = ep->info.width;
    ep->ph = ep->info.height;

    if (status != PixmapSuccess)
    {
	char buf[1024];
	switch (status)
	{
	    case PixmapOpenFailed:
		sprintf(buf, "Cannot Open Pixmap File\n'%s'", filename);
		break;
	    case PixmapFileInvalid:
		sprintf(buf, "Invalid Pixmap File\n'%s'", filename);
		break;
	    case PixmapNoMemory:
		sprintf(buf, "Out of Memory Opening Pixmap File\n'%s'",
		    filename);
		break;
	    case PixmapColorFailed:
		sprintf(buf, "No Colors Found In Pixmap File\n'%s'", filename);
		break;
	    case PixmapColorError:
		sprintf(buf, 
		    "Cannot Allocate Enough Color Cells\nPixmap File\n'%s'",
		    filename);
		break;
	}
	popupMessage(XPE_Error, buf, NULL, NULL, NULL);
	ep->pw = appData.pixmapWidth;
	ep->ph = appData.pixmapHeight;
	createPixmap(ep);
	return;
    }
    /* update our colormap info */
    newLoadedPixels(ep->info.pixels, ep->info.npixels, 
	ep->info.ncolors, ep->info.colorTable, ep->info.pixels);

    ep->workpixmap = XCreatePixmap(display, 
	RootWindowOfScreen(XtScreenOfObject(ep->shell)), ep->pw, ep->ph,
#ifdef WIDE_WORK_PIXMAP
	DefaultDepth(display, screen));
#else
	1);
#endif

    ep->image = XGetImage(display, ep->pixmap, 0, 0, ep->pw, ep->ph, AllPlanes, ZPixmap);
    ep->workimage = XSubImage(ep->image, 0, 0, ep->pw, ep->ph);
    ep->undoimage = XSubImage(ep->image, 0, 0, ep->pw, ep->ph);

    FirstArg(XmNwidth, (ep->pw * (ep->scale + 1)) + 1);
    NextArg(XmNheight, (ep->ph * (ep->scale + 1)) + 1);
    XtSetValues(ep->workarea, Args, ArgCount);
    FirstArg(XmNwidth, ep->pw);
    NextArg(XmNheight, ep->ph);
    XtSetValues(ep->pixarea, Args, ArgCount);

    ep->pixfile = XtNewString(filename);
}

#ifdef _NO_PROTO
editorInit()
#else
editorInit()
#endif
{
extern void selectPrimCB();
extern void clearCB();
extern void undoCB();
extern void saveSnapshotCB();
extern void restoreSnapshotCB();
extern void compressColorsCB();
extern void loadBitmapCB();
extern void helpMsgCB();
extern void textSelectCB();
extern void textDoneCB();
extern void symNameCB();

    WcRegisterCallback(appCon, "redrawImageCB", 
	(XtCallbackProc)redrawImageCB, NULL);
    WcRegisterCallback(appCon, "setLabelPixmapCB", 
	(XtCallbackProc)setLabelPixmapCB, NULL);
    WcRegisterCallback(appCon, "selectPrimCB", 
	(XtCallbackProc)selectPrimCB, NULL);
    WcRegisterCallback(appCon, "clearCB", 
	(XtCallbackProc)clearCB, NULL);
    WcRegisterCallback(appCon, "undoCB", 
	(XtCallbackProc)undoCB, NULL);
    WcRegisterCallback(appCon, "saveSnapshotCB", 
	(XtCallbackProc)saveSnapshotCB, NULL);
    WcRegisterCallback(appCon, "restoreSnapshotCB", 
	(XtCallbackProc)restoreSnapshotCB, NULL);
    WcRegisterCallback(appCon, "compressColorsCB", 
	(XtCallbackProc)compressColorsCB, NULL);
    WcRegisterCallback(appCon, "saveCB", 
	(XtCallbackProc)saveCB, NULL);
    WcRegisterCallback(appCon, "saveAsCB", 
	(XtCallbackProc)saveAsCB, NULL);
    WcRegisterCallback(appCon, "openCB", 
	(XtCallbackProc)openCB, NULL);
    WcRegisterCallback(appCon, "newCB", 
	(XtCallbackProc)newCB, NULL);
    WcRegisterCallback(appCon, "exitCB", 
	(XtCallbackProc)exitCB, NULL);
    WcRegisterCallback(appCon, "closeCB", 
	(XtCallbackProc)closeCB, NULL);
    WcRegisterCallback(appCon, "fontCB", 
	(XtCallbackProc)fontCB, NULL);
    WcRegisterCallback(appCon, "loadBitmapCB", 
	(XtCallbackProc)loadBitmapCB, NULL);
    WcRegisterCallback(appCon, "editInfoCB", 
	(XtCallbackProc)editInfoCB, NULL);
    WcRegisterCallback(appCon, "fillToBoundaryCB", 
	(XtCallbackProc)fillToBoundaryCB, NULL);
    WcRegisterCallback(appCon, "helpMsgCB", 
	(XtCallbackProc)helpMsgCB, NULL);
    WcRegisterCallback(appCon, "textSelectCB", 
	(XtCallbackProc)textSelectCB, NULL);
    WcRegisterCallback(appCon, "textDoneCB", 
	(XtCallbackProc)textDoneCB, NULL);
    WcRegisterCallback(appCon, "symNameCB", 
	(XtCallbackProc)symNameCB, NULL);
}

#ifdef _NO_PROTO
void editorCreate(parent, filename)
Widget parent; String filename;
#else
void editorCreate(Widget parent, String filename)
#endif
{
    EditorPtr ep;
    DeclareArgs(10);
    Widget wid;

    ep = (EditorPtr) XtCalloc(1, sizeof(Editor));

    ep->fillToBoundary = 0;
    ep->bitmapimage = NULL;
    ep->textimage = NULL;
    ep->clipboardPixmap = 0;
    ep->clipID = 0;
    ep->scrollInProgress = 0;
    ep->snapshotimage = 0;
    ep->scale = appData.scale;
    ep->buttonPressFunction = NULL;
    ep->enteredText = NULL;

    ep->shell = XtCreatePopupShell("editor", topLevelShellWidgetClass, 
	parent, NULL, 0);
    WcWidgetCreation(ep->shell);

    wid = WcFullNameToWidget(ep->shell, ".mainWindow");
    FirstArg(XmNuserData, ep);
    XtSetValues(wid, Args, ArgCount);

    ep->gc = XCreateGC(display, 
	RootWindowOfScreen(XtScreenOfObject(parent)), 0, NULL);
#ifdef WIDE_WORK_PIXMAP
    ep->workgc = XCreateGC(display,
	RootWindowOfScreen(XtScreenOfObject(parent)), 0, NULL);
#else
    ep->workpixmap = XCreatePixmap(display, 
	RootWindowOfScreen(XtScreenOfObject(parent)), 1, 1, 1);
    ep->workgc = XCreateGC(display, ep->workpixmap, 0, NULL);
    XFreePixmap(display, ep->workpixmap);
    ep->workpixmap = NULL;
#endif

    ep->workScrollWin = WcFullNameToWidget(ep->shell, 
		".mainWindow.pane.zoomSW");

    ep->workarea = WcFullNameToWidget(ep->workScrollWin, 
		".ScrolledWindowClipWindow.zoom");

    ep->pixarea = WcFullNameToWidget(ep->shell, 
		".mainWindow.pane.normalSW.ScrolledWindowClipWindow.normal");

    ep->messageArea = WcFullNameToWidget(ep->shell, 
		".mainWindow.message");

    ep->textEntry = WcFullNameToWidget(ep->workarea, ".textEntry");

    FirstArg(XmNclipWindow, &ep->clipWidget);
    XtGetValues(ep->workScrollWin, Args, ArgCount);

    addEventHandler(ep->workarea, ep);
    addEventHandler(ep->pixarea, ep);

    FirstArg(XmNbackground, &ep->workbg);
    NextArg(XmNforeground, &ep->workfg);
    XtGetValues(ep->workarea, Args, ArgCount);

    if (filename)
    {
	readPixmap(ep, filename);
	FirstArg(XmNtitle, filename);
	XtSetValues(ep->shell, Args, ArgCount);
	sayMessage(ep, filename);
    }
    else
    {
	ep->pw = appData.pixmapWidth;
	ep->ph = appData.pixmapHeight;
	createPixmap(ep);
	FirstArg(XmNtitle, "New Pixmap");
	XtSetValues(ep->shell, Args, ArgCount);
	sayMessage(ep, "New Pixmap");
    }

    XtPopup(ep->shell, XtGrabNone);
}

#ifdef _NO_PROTO
editorDestroy(ep)
EditorPtr ep;
#else
editorDestroy(EditorPtr ep)
#endif
{
    XtPopdown(ep->shell);
    XFreeGC(display, ep->gc);
    XFreeGC(display, ep->workgc);

    XFreePixmap(display, ep->pixmap);
    XFreePixmap(display, ep->workpixmap);

    if (ep->enteredText)
	XtFree(ep->enteredText);

    if (ep->clipboardPixmap)
	XFreePixmap(display, ep->clipboardPixmap);
    if (ep->textimage)
	XDestroyImage(ep->textimage);
    if (ep->bitmapimage)
	XDestroyImage(ep->bitmapimage);

    if (ep->clipID)
    {
	while (XmClipboardUndoCopy(display, XtWindow(ep->workarea))
		== ClipboardLocked)
	    sleep(1);
    }
    XDestroyImage(ep->image);
    XDestroyImage(ep->workimage);
    XDestroyImage(ep->undoimage);
    if (ep->snapshotimage)
	XDestroyImage(ep->snapshotimage);

    XtFree(ep->pixfile);

    XtDestroyWidget(ep->shell);

    XtFree((XtPointer)ep);
}
