/*
 * font editor routines
 */

#include "declare.h"
#include "Argtricks.h"
#include "ctype.h"
#include <Xm/PushBG.h>
#include <Xm/ToggleB.h>
#include <Xm/DialogS.h>

#include <X11/Wc/WcCreate.h>

#define FONT_MASK	"-*-iso8859-1"
#define MAX_NAMES	1000

typedef struct _string_list {
    String 		s;
    struct _string_list	*next;
} StringListItem, *StringList;

typedef struct _font_info {
    int		size;
    Boolean	exists;		/* does size exists? */
    Boolean	slant_reg, slant_ital, slant_obl;
    Boolean	weight_med, weight_bold;
} FontInfo;

static Widget fontShell = NULL;	/* main shell */

FontInfo *fontInfo = NULL;	/* one per sortSize entry */

StringList fonts = NULL;
StringList family = NULL;
StringList size = NULL;
StringList sortSize = NULL;	/* actually in reverse order!! */
int	numSizes = 0;

enum { SLANT_REGULAR, SLANT_ITALIC, SLANT_OBLIQUE };
enum { WEIGHT_MEDIUM, WEIGHT_BOLD};

int currentSize;	/* current size selected */
int currentSlant;	/* current slant selected */
int currentWeight; 	/* current weight selected */
String currentFamily;	/* current font family selected */
String firstFamily;	/* first family of fonts selected */

XFontStruct *currentFont = NULL;

#ifdef _NO_PROTO
static String parseDash(s)
String s;
#else
static String parseDash(String s)
#endif
{
    while (*s)
    {
	if (*s == '-')
	{
	    *s = 0;
	    return(++s);
	}
	s++;
    }
    return(NULL);
}

#ifdef _NO_PROTO
static void addList(list, value)
StringList *list; String value;
#else
static void addList(StringList *list, String value)
#endif
{
    StringList l = *list;
    StringListItem *newItem;
    while (l)
    {
	if (strcmp(l->s, value) == 0)
	    return;
	l = l->next;
    }
    newItem = (StringListItem *)malloc(sizeof(StringListItem));
    newItem->s = (String )malloc(strlen(value)+1);
    strcpy(newItem->s, value);
    newItem->next = *list;
    *list = newItem;
}

#ifdef _NO_PROTO
static void addFont(name)
String name;
#else
static void addFont(String name)
#endif
{
    char scratch[1024];		/* hope this is always big enough */
    char namecopy[1024];	/* ditto */
    String Ffamily;
    String Fname;
    String Fweight;
    String Fslant;
    String FsWidth;
    String Fadstyle;
    String FpixelSize;
    String FpointSize;
#ifdef DONT_CARE_ABOUT_THESE
    String FresX;
    String FresY;
    String Fspacing;
    String FavgWidth;
    String Fregistry;
    String Fencoding;
#endif

    strcpy(namecopy, name);
    Ffamily	= parseDash(namecopy);
    Fname	= parseDash(Ffamily);
    Fweight	= parseDash(Fname);
    Fslant	= parseDash(Fweight);
    FsWidth	= parseDash(Fslant);
    Fadstyle	= parseDash(FsWidth);
    FpixelSize	= parseDash(Fadstyle);
    FpointSize	= parseDash(FpixelSize);
#ifdef DONT_CARE_ABOUT_THESE
    FresX	= parseDash(FpointSize);
    FresY	= parseDash(FresX);
    Fspacing	= parseDash(FresY);
    FavgWidth	= parseDash(Fspacing);
    Fregistry	= parseDash(FavgWidth);
    Fencoding	= parseDash(Fregistry);
#endif

    /* rebuild the font name */
    strcpy(scratch, "-*-");
    strcat(scratch, Fname);
    strcat(scratch, "-");
    strcat(scratch, Fweight);
    strcat(scratch, "-");
    strcat(scratch, Fslant);
    strcat(scratch, "-*--");
    strcat(scratch, FpixelSize);
    strcat(scratch, "-*");

    /* add names, sizes and original names to list */
    addList(&family, Fname);
    addList(&fonts, scratch);
    addList(&size, FpixelSize);
}

#ifdef _NO_PROTO
static int countList(list)
StringList list;
#else
static int countList(StringList list)
#endif
{
    int count = 0;
    while (list)
    {
	count++;
	list = list->next;
    }
    return(count);
}

#ifdef _NO_PROTO
void selectFont()
#else
void selectFont()
#endif
{
    Widget w;
    void createFontEditor();
    if (!fontShell)
	createFontEditor();
    w = WcFullNameToWidget(topLevel, "*fontForm");

    /* be sure that we are visible */
    if (XtIsRealized(topLevel))
	XtMapWidget(topLevel);

    XtManageChild(w);
}

#ifdef _NO_PROTO
static Boolean matchFamily(name, fam)
String name; String fam;
#else
static Boolean matchFamily(String name, String fam)
#endif
{
    name += 3;		/* skip "-*-" */
    while (*name && *fam)
    {
	if (*name++ != *fam++)
	    return(False);
    }
    /* we should now be at next "-" in the full font name, if not... */
    return(*name == '-');
}

Widget wSizeList = NULL, wSize, wMedium, wBold, 
	wRegular, wItalic, wOblique, wSample;

/*
 * load font into server
 */
#ifdef _NO_PROTO
static void loadFont()
#else
static void loadFont()
#endif
{
    char name[256];
    String wt, sl;

    wt = (currentWeight == WEIGHT_MEDIUM ? "medium" : "bold");
    if (currentSlant == SLANT_ITALIC)
	sl = "i";
    else if (currentSlant == SLANT_OBLIQUE)
	sl = "o";
    else
	sl = "r";
    sprintf(name, "-*-%s-%s-%s-*-%d-*", currentFamily, wt, sl, currentSize);
    if (currentFont)
	XFreeFont(display, currentFont);

    currentFont = XLoadQueryFont(display, name);
    if (!currentFont)
    {
	fprintf(stderr, "font load failed on font '%s'!\n", name);
	fprintf(stderr, "loading 'fixed' font\n");
	currentFont = XLoadQueryFont(display, "fixed");
	if (!currentFont)
	{
	    fprintf(stderr, "cannot load 'fixed' font!\n");
	    exit(-1);
	}
    }
    if (XtIsRealized(wSample))
	XClearArea(display, XtWindow(wSample), 0, 0, 0, 0, True);
}

/*
 * update the available (sensitive) widgets
 */
#ifdef _NO_PROTO
static void updateSelectionInfo(sizes)
Boolean sizes;
#else
static void updateSelectionInfo(Boolean sizes)
#endif
{
    int i;
    FontInfo *fi;
    Widget wid, firstSize = NULL;
    DeclareArgs(2);
    WidgetList wl;
    Cardinal nc;

    /* get widget IDs first time */
    if (!wSizeList)
    {
	wSizeList = XtNameToWidget(topLevel, "*fontSizeMenu");
	wSize = XtNameToWidget(topLevel, "*fontSize");
	wMedium = XtNameToWidget(topLevel, "*medium");
	wBold = XtNameToWidget(topLevel, "*bold");
	wRegular = XtNameToWidget(topLevel, "*regular");
	wItalic = XtNameToWidget(topLevel, "*italic");
	wOblique = XtNameToWidget(topLevel, "*oblique");
	wSample = XtNameToWidget(topLevel, "*fontSample");
    }
    FirstArg(XmNchildren, &wl);
    NextArg(XmNnumChildren, &nc);
    XtGetValues(wSizeList, Args, ArgCount);

    /* sanity check */
    if (nc != numSizes)
    {
	fprintf(stderr, "I'm confused:  nc=%d   expecting %d\n", nc, numSizes);
	exit(-1);
    }

    /* set sizes sensitivity */
    if (sizes)
    {
	for (i=0; i<numSizes; i++)
	{
	    wid = wl[numSizes - i - 1];
	    fi = fontInfo + i;
	    XtSetSensitive(wid, fi->exists);
	    if (fi->exists)
		firstSize = wid;
	}
    }
    FirstArg(XmNmenuHistory, &wid);
    XtGetValues(wSize, Args, ArgCount);

    /* is selected size if not valid for this family then change  */
    if (!wid || !XtIsSensitive(wid))
    {
	FirstArg(XmNmenuHistory, firstSize);
	XtSetValues(wSize, Args, ArgCount);
	wid = firstSize;
    }
    /* go from selected size to fontInfo field */
    for (i=0; i<numSizes; i++)
	if (wid == wl[numSizes - i - 1])
	    break;

    fi = fontInfo + i;
    if (currentSize != fi->size)
	currentSize = fi->size;

    XtSetSensitive(wMedium, fi->weight_med);
    XtSetSensitive(wBold, fi->weight_bold);

    /* invalid toggle button selected? */
    if (!fi->weight_bold)
    {
	XmToggleButtonSetState(wBold, False, False);
	XmToggleButtonSetState(wMedium, True, False);
	currentWeight = WEIGHT_MEDIUM;
    }

    XtSetSensitive(wRegular, fi->slant_reg);
    XtSetSensitive(wItalic, fi->slant_ital);
    XtSetSensitive(wOblique, fi->slant_obl);

    /* invalid toggle button selected? */
    if (currentSlant == SLANT_ITALIC && !fi->slant_ital)
    {
	XmToggleButtonSetState(wItalic, False, False);
	XmToggleButtonSetState(wRegular, True, False);
	currentSlant = SLANT_REGULAR;
    }
    if (currentSlant == SLANT_OBLIQUE && !fi->slant_obl)
    {
	XmToggleButtonSetState(wOblique, False, False);
	XmToggleButtonSetState(wRegular, True, False);
	currentSlant = SLANT_REGULAR;
    }
    loadFont();
}

/* ARGSUSED */
#ifdef _NO_PROTO
static void sizeCB(wid, size, unused)
Widget wid; int size; XtPointer unused;
#else
static void sizeCB(Widget wid, int size, XtPointer unused)
#endif
{
    if (currentSize != size)
    {
	currentSize = size;
	updateSelectionInfo(False);
    }
}

#ifdef _NO_PROTO
static void setFontInfo(fam)
String fam;
#else
static void setFontInfo(String fam)
#endif
{
    StringList f;
    int i, sz;
    String s;

    f = fonts;
    while (f)
    {
	if (matchFamily(f->s, fam))
	{
	    /* get font name:  -*-<family>-<weight>-<slant>-*--<size>-*- */
	    /* we can't use parse dash since it inserts nulls into string */
	    s = f->s;

	    /* skip to size field, it's after "--" */
	    while (*s != '-' || *(s+1) != '-')
		s++;

	    /* size field should be next */
	    s += 2;
	    sz = atoi(s);

	    /* find matching fontInfo struct */
	    for (i=0; i<numSizes; i++)
	    {
		if (fontInfo[i].size == sz)
		    break;
	    }
	    if (i == numSizes)
	    {
		fprintf(stderr, "Bad size in '%s'\n", f->s);
		exit(-1);
	    }

	    /* now lets go back and look for weight and slant */
	    s = f->s;
	    s += 3;

	    /* skip family field */
	    while (*s != '-')
		s++;

	    /* we should be at weight */
	    s++;
	    if (*s == 'm')
		fontInfo[i].weight_med = True;
	    else if (*s == 'b')
		fontInfo[i].weight_bold = True;

	    /* now go to slant */
	    while (*s != '-')
		s++;
	    s++;
	    if (*s == 'r')
		fontInfo[i].slant_reg = True;
	    else if (*s == 'i')
		fontInfo[i].slant_ital = True;
	    else if (*s == 'o')
		fontInfo[i].slant_obl = True;

	    fontInfo[i].exists = True;
	}
	f = f->next;
    }
    updateSelectionInfo(True);
}

/* ARGSUSED */
#ifdef _NO_PROTO
static void familyCB(wid, fam, unused)
Widget wid; String fam; XtPointer unused;
#else
static void familyCB(Widget wid, String fam, XtPointer unused)
#endif
{
    int i;
    if (fam == currentFamily)
	return;
    for (i=0; i<numSizes; i++)
    {
	fontInfo[i].exists = False;
	fontInfo[i].slant_reg = fontInfo[i].slant_ital =
	    fontInfo[i].slant_obl = False;
	fontInfo[i].weight_med = fontInfo[i].weight_bold = False;
    }
    currentFamily = fam;
    setFontInfo(fam);
}

/* ARGSUSED */
#ifdef _NO_PROTO
static void slantCB(wid, slant, cb)
Widget wid; String slant;
	XmToggleButtonCallbackStruct *cb;
#else
static void slantCB(Widget wid, String slant,
	XmToggleButtonCallbackStruct *cb)
#endif
{
    int oldSlant = currentSlant;
    if (cb->set)
    {
	if (strcmp(slant, "r") == 0)
	    currentSlant = SLANT_REGULAR;
	else if (strcmp(slant, "i") == 0)
	    currentSlant = SLANT_ITALIC;
	else
	    currentSlant = SLANT_OBLIQUE;
    }
    if (oldSlant != currentSlant)
	loadFont();
}

/* ARGSUSED */
#ifdef _NO_PROTO
static void weightCB(wid, weight, cb)
Widget wid; String weight; XmToggleButtonCallbackStruct *cb;
#else
static void weightCB(Widget wid, String weight,
	XmToggleButtonCallbackStruct *cb)
#endif
{
    int oldWeight = currentWeight;
    if (cb->set)
    {
	if (strcmp(weight, "m") == 0)
	    currentWeight = WEIGHT_MEDIUM;
	else
	    currentWeight = WEIGHT_BOLD;
    }
    if (oldWeight != currentWeight)
	loadFont();
}

#ifdef _NO_PROTO
static int compareLists(a, b)
StringList *a; StringList *b;
#else
static int compareLists(StringList *a, StringList *b)
#endif
{
    int ia, ib;

    if (isalpha((*a)->s[0]) || isalpha((*b)->s[0]))
	return(strcmp((*a)->s, (*b)->s));

    ia = atoi((*a)->s);
    ib = atoi((*b)->s);
    if (ia < ib)
	return(-1);
    else if (ia == ib)
	return(0);
    else
	return(1);
}

#ifdef _NO_PROTO
typedef	int (*CompareFunc)();
#else
typedef	int (*CompareFunc)(const void *, const void *);
#endif

#ifdef _NO_PROTO
static void buildMenu(parent, list, addToSortSize)
Widget parent; StringList list; int addToSortSize;
#else
static void buildMenu(Widget parent, StringList list, int addToSortSize)
#endif
{
    int count, i;
    StringList *ptrs, l;
    DeclareArgs(1);
    XmString xms;
    Widget wid;

    count = countList(list);
    ptrs = (StringList *) malloc(count * sizeof(StringList));
    l = list;
    for (i=0; i<count; i++)
    {
	ptrs[i] = l;
	l = l->next;
    }

    qsort(ptrs, count, sizeof(StringList), (CompareFunc)compareLists);

    for (i=0; i<count; i++)
    {
	l = ptrs[i];
	xms = XmStringCreateSimple(l->s);

	FirstArg(XmNlabelString, xms);
	wid = XmCreatePushButtonGadget(parent, "ob", Args, ArgCount);
	if (addToSortSize)
	    XtAddCallback(wid, XmNactivateCallback,
			(XtCallbackProc)sizeCB, (XtPointer)atoi(l->s));
	else
	    XtAddCallback(wid, XmNactivateCallback, 
			(XtCallbackProc)familyCB, (XtPointer)l->s);
	XtManageChild(wid);
	XmStringFree(xms);

	/* this adds them in reverse order since it insert's at start */
	if (addToSortSize)
	    addList(&sortSize, l->s);
    }
    if (!addToSortSize)
	firstFamily = (*ptrs)->s;
    free(ptrs);
}

/* ARGSUSED */
#ifdef _NO_PROTO
void buildFontMenu(wid, menuName, unused)
Widget wid; String menuName; XtPointer unused;
#else
void buildFontMenu(Widget wid, String menuName, XtPointer unused)
#endif
{
    int i;
    StringList l;

    if (strcmp(menuName, "family") == 0)
    {
	buildMenu(wid, family, False);
    }
    else
    {
	buildMenu(wid, size, True);
	currentSize = atoi(sortSize->s);
	numSizes = countList(size);
	fontInfo = (FontInfo *) malloc(numSizes * sizeof(FontInfo));

	l = sortSize;
	for (i=0; i<numSizes; i++)
	{
	    fontInfo[i].exists = False;
	    fontInfo[i].size = atoi(l->s);
	    fontInfo[i].slant_reg = fontInfo[i].slant_ital =
		fontInfo[i].slant_obl = False;
	    fontInfo[i].weight_med = fontInfo[i].weight_bold = False;
	    l = l->next;
	}
    }
    currentSlant = SLANT_REGULAR;
    currentWeight = WEIGHT_MEDIUM;
}

/* ARGSUSED */
#ifdef _NO_PROTO
static void showFont(wid, unused, cb)
Widget wid; XtPointer unused; 
	XmDrawingAreaCallbackStruct *cb;
#else
static void showFont(Widget wid, XtPointer unused, 
	XmDrawingAreaCallbackStruct *cb)
#endif
{
    static GC gc;
    Window win = cb->window;
    Pixel fg;
    DeclareArgs(1);

    if (!(XtIsRealized(wid)) || !currentFont)
	return;

    if (!gc)
    {
	gc = XCreateGC(display, win, NULL, 0);

	FirstArg(XmNforeground, &fg);
	XtGetValues(wid, Args, ArgCount);
	XSetForeground(display, gc, fg);
    }
    XSetFont(display, gc, currentFont->fid);
    XDrawString(display, win, gc, 0, currentFont->ascent, 
	"ABCDEFGHIJKLMNOPQRSTUVWXYZ", 26);
    XDrawString(display, win, gc, 0, 
	currentFont->ascent + currentFont->descent + currentFont->ascent, 
	"abcdefghijklmnopqrstuvwxyz", 26);
}

#ifdef _NO_PROTO
void initFontEditor()
#else
void initFontEditor()
#endif
{
    currentFont = XLoadQueryFont(display, "fixed");
    if (!currentFont)
    {
	fprintf(stderr, "cannot load 'fixed' font!\n");
	exit(-1);
    }
}

#ifdef _NO_PROTO
void createFontEditor()
#else
void createFontEditor()
#endif
{
String *fontnames, *orig;
int count;
Widget menu, wid, form;

    orig = fontnames = XListFonts(display, FONT_MASK, MAX_NAMES, &count);

    while (count-- > 0)
	addFont(*fontnames++);

    XFreeFontNames(orig);

    fontShell = XmCreateDialogShell(topLevel, "fontInfo", NULL, 0);
    WcRegisterCallback(appCon, "buildFontMenu", 
		(XtCallbackProc)buildFontMenu, NULL);
    WcRegisterCallback(appCon, "slantCB",
		(XtCallbackProc)slantCB, NULL);
    WcRegisterCallback(appCon, "weightCB",
		(XtCallbackProc)weightCB, NULL);
    WcRegisterCallback(appCon, "showFont",
		(XtCallbackProc)showFont, NULL);
    WcWidgetCreation(fontShell);

    /* update font info for first selection */
    setFontInfo(currentFamily = firstFamily);

}
