/*
 * xpe - x pimap editor
 */
#include "declare.h"
#include <X11/Shell.h>
#include <X11/StringDefs.h>
#include <Xm/Xm.h>
#include "Argtricks.h"
#include <Xm/Label.h>

#include <X11/Wc/WcCreate.h>

#define MY_CLASS	"XPE"

XtAppContext appCon;
Display *display;
int screen;
Widget topLevel;
Pixel curPixel = 0;
Colormap cmap;

/* application resources */
AppData appData;

#define XtNpixmapWidth		"pixmapWidth"
#define XtCPixmapWidth		"PixmapWidth"
#define XtNpixmapHeight		"pixmapHeight"
#define XtCPixmapHeight		"PixmapHeight"
#define XtNscale		"scale"
#define XtCScale		"Scale"
#define XtNemptyCmap		"emptyCmap"
#define XtCEmptyCmap		"EmptyCmap"
#define XtNcolorFile		"colorFile"
#define XtCColorFile		"ColorFile"

static XtResource resources[] = {
{ XtNpixmapWidth, XtCPixmapWidth, XtRInt, sizeof(int),
	XtOffset(AppDataPtr, pixmapWidth), XtRImmediate, (caddr_t)32},
{ XtNpixmapHeight, XtCPixmapHeight, XtRInt, sizeof(int),
	XtOffset(AppDataPtr, pixmapHeight), XtRImmediate, (caddr_t)32},
{ XtNfile, XtCFile, XtRString, sizeof(String),
	XtOffset(AppDataPtr, loadfile), XtRImmediate, (caddr_t)0},
{ XtNscale, XtCScale, XtRInt, sizeof(int),
	XtOffset(AppDataPtr, scale), XtRImmediate, (caddr_t)8},
{ XtNemptyCmap, XtCEmptyCmap, XtRBool, sizeof(Bool),
	XtOffset(AppDataPtr, emptyCmap), XtRImmediate, (caddr_t)False},
{ XtNcolorFile, XtCColorFile, XtRString, sizeof(String),
	XtOffset(AppDataPtr, colorfile), XtRImmediate, (caddr_t)0},
};

XrmOptionDescRec options[] = {
    {"-width",	"*pixmapWidth",		XrmoptionSepArg, 0},
    {"-height",	"*pixmapHeight",	XrmoptionSepArg, 0},
    {"-file",	"*file",		XrmoptionSepArg, 0},
    {"-scale",	"*scale",		XrmoptionSepArg, 0},
    {"-emptycmap",	"*emptyCmap",	XrmoptionNoArg, (caddr_t)"True"},
    {"-colorfile",	"*colorFile",	XrmoptionSepArg, 0},
};

/* */
/* usage error */
/* */

#ifdef _NO_PROTO
void Usage(name)
char *name;
#else
void Usage(char *name)
#endif
{
    printf("Usage: %s [options...]\n", name);
    printf("\
 -file   name		file name to edit\n\
 -width  w		sets the desired pixmap width\n\
 -height h		sets the desired pixmap height\n\
 -scale  s		sets the pixel scale size\n\
 -emptycmap		do not color root colormap pixels\n\
 -colorfile name	file of color names to initialize colormap with\n\
");
    exit(-1);
}


#ifdef _NO_PROTO
main(argc, argv)
int argc; char **argv;
#else
main(int argc, char **argv)
#endif
{
DeclareArgs(1);
#ifdef _NO_PROTO
extern void MriRegisterMotif(); 
#else
extern void MriRegisterMotif(XtAppContext context); 
#endif

    /* got to do this */
    XtToolkitInitialize();

    /* only use one app context */
    appCon = XtCreateApplicationContext();

    display = XtOpenDisplay(appCon, NULL, argv[0], MY_CLASS, 
	options, XtNumber(options), (unsigned int *)&argc, argv);
    screen = DefaultScreen(display);
    if (DisplayCells(display, screen) != 256)
    {
	printf("This program is hardcoded for a 256 color display\n");
	printf("I din't think this is one.  Sorry\n");
	exit(-1);
    }

    if (argc > 1)
	Usage(argv[0]);
    
    /* initialize Wcl */
    MriRegisterMotif(appCon);

    /* create the master root shell */
    topLevel = XtAppCreateShell("xpe", MY_CLASS,
	applicationShellWidgetClass, display, NULL, 0);
    
    /* get our application resources */
    XtGetApplicationResources(topLevel, (XtPointer)&appData,
	resources, XtNumber(resources), NULL, 0);

    /* be sure Motif gets its converters loaded first */
    XtInitializeWidgetClass(xmLabelWidgetClass);

    /* install special name to widget converters */
    installConverters(appCon);
    
    createColormap();

    FirstArg(XmNcolormap, cmap);
    XtSetValues(topLevel, Args, ArgCount);

    initColorWindow();
    editorInit();
    initFontEditor();

    createColorWindow(topLevel);

    editorCreate(topLevel, appData.loadfile);

    updateColormap();
    XtRealizeWidget(topLevel);
    XtAppMainLoop(appCon);
}
