/*
 * editor window class routines - primitive functions
 */

#include "declare.h"

#include "Argtricks.h"
#include <X11/cursorfont.h>
#include <Xm/Xm.h>
#include <Xm/CutPaste.h>
#include <Xm/ScrollBar.h>
#include <Xm/TextF.h>

#include "editor.h"

#define MAX_PTS	1000

#define SCROLL_LEFT	1
#define SCROLL_RIGHT	2
#define SCROLL_UP	4
#define SCROLL_DOWN	8

/* scroll delay in milliseconds */
#define SCROLL_INTERVAL	100

#ifdef _NO_PROTO
EditorPtr getEditor();
#else
EditorPtr getEditor(Widget wid);
#endif

/*
** remove duplicate colors
*/
/* ARGSUSED */
#ifdef _NO_PROTO
void compressColorsCB(wid, unused, unused_too)
Widget wid; XtPointer unused; XtPointer unused_too;
#else
void compressColorsCB(Widget wid, XtPointer unused, XtPointer unused_too)
#endif
{
int c, newc, x, y;
int compressed[256];
EditorPtr ep;

    ep = getEditor(wid);

    for (x=0; x<256; x++)
	compressed[x] = -1;
    for (x=0; x < ep->pw; x++)
    {
	for (y=0; y<ep->ph; y++)
	{
	    c = (int)XGetPixel(ep->image, x, y);
	    if (compressed[c] == -1)
	    {
		newc = compressColor(c);
		compressed[c] = newc;
	    }
	    XPutPixel(ep->image, x, y, compressed[c]);
	}
    }
    XPutImage(display, ep->pixmap, ep->gc, ep->image, 0, 0, 0, 0, ep->pw, ep->ph);
    redrawWidget(ep->workarea);
    redrawWidget(ep->pixarea);
}

/*
** clear the pixmap data to zero
*/
/* ARGSUSED */
#ifdef _NO_PROTO
void clearCB(wid, unused, unused_too)
Widget wid; XtPointer unused; XtPointer unused_too;
#else
void clearCB(Widget wid, XtPointer unused, XtPointer unused_too)
#endif
{
int x, y;
EditorPtr ep;

    ep = getEditor(wid);
    for (x=0; x<ep->pw; x++)
	for (y=0; y<ep->ph; y++)
	    XPutPixel(ep->image, x, y, 0);
    XSetForeground(display, ep->gc, 0);
    XFillRectangle(display, ep->pixmap, ep->gc, 0, 0, ep->pw, ep->ph);

    redrawWidget(ep->workarea);
    redrawWidget(ep->pixarea);
}

/*
** save current state for undo
*/
#ifdef _NO_PROTO
void saveState(ep)
EditorPtr ep;
#else
void saveState(EditorPtr ep)
#endif
{
int x, y;
Pixel p;

    for (x=0; x<ep->pw; x++)
    {
	for (y=0; y<ep->ph; y++)
	{
	    p = XGetPixel(ep->image, x, y);
	    XPutPixel(ep->undoimage, x, y, p);
	}
    }
}

/*
** save a snapshot of the current image
*/
/* ARGSUSED */
#ifdef _NO_PROTO
void saveSnapshotCB(wid, unused, unused_too)
Widget wid; XtPointer unused; XtPointer unused_too;
#else
void saveSnapshotCB(Widget wid, XtPointer unused, XtPointer unused_too)
#endif
{
EditorPtr ep;

    ep = getEditor(wid);
    if (ep->snapshotimage)
	XDestroyImage(ep->snapshotimage);
    ep->snapshotimage = XSubImage(ep->image, 0, 0, ep->pw, ep->ph);
}

/*
** restore data from the snapshot image
*/
/* ARGSUSED */
#ifdef _NO_PROTO
void restoreSnapshotCB(wid, unused, unused_too)
Widget wid; XtPointer unused; XtPointer unused_too;
#else
void restoreSnapshotCB(Widget wid, XtPointer unused, XtPointer unused_too)
#endif
{
int x, y;
Pixel p;
EditorPtr ep;

    ep = getEditor(wid);
    if (!ep->snapshotimage)
	return;
    saveState(ep);
    for (x=0; x<ep->pw; x++)
	for (y=0; y<ep->ph; y++)
	{
	    p = XGetPixel(ep->snapshotimage, x, y);
	    XPutPixel(ep->image, x, y, p);
	}
    XPutImage(display, ep->pixmap, ep->gc, ep->image, 0, 0, 0, 0, ep->pw, ep->ph);
    redrawWidget(ep->workarea);
    redrawWidget(ep->pixarea);
}

/*
** restore data from the undo image
*/
/* ARGSUSED */
#ifdef _NO_PROTO
void undoCB(wid, unused, unused_too)
Widget wid; XtPointer unused; XtPointer unused_too;
#else
void undoCB(Widget wid, XtPointer unused, XtPointer unused_too)
#endif
{
XImage *im;
EditorPtr ep;

    ep = getEditor(wid);
    im = ep->image;
    ep->image = ep->undoimage;
    ep->undoimage = im;
    XPutImage(display, ep->pixmap, ep->gc, ep->image, 0, 0, 0, 0, ep->pw, ep->ph);
    redrawWidget(ep->workarea);
    redrawWidget(ep->pixarea);
}

/*
** handle configure and visibility events in the zoom window and clip window
*/
/* ARGSUSED */
#ifdef _NO_PROTO
void configInputEH(wid, ep, e)
Widget wid; EditorPtr ep; XEvent *e;
#else
void configInputEH(Widget wid, EditorPtr ep, XEvent *e)
#endif
{
int dummy;
unsigned int udummy;
Window root;

    if (e->type == ConfigureNotify || e->type == MapNotify)
    {
	XGetGeometry(display, XtWindow(ep->clipWidget), &root, &dummy, &dummy,
		(unsigned int *)&ep->clipWinW, (unsigned int *)&ep->clipWinH,
		&udummy, &udummy);
	XGetGeometry(display, XtWindow(ep->workarea), &root,
		&ep->workWinX, &ep->workWinY,
		(unsigned int *)&ep->workWinW, (unsigned int *)&ep->workWinH,
		&udummy, &udummy);
    }
    else if (e->type == VisibilityNotify)
    {
	ep->workObscured = 
		(e->xvisibility.state == VisibilityPartiallyObscured);
    }
}

/*
** handle auto scrolling of work window
*/
/* ARGSUSED */
#ifdef _NO_PROTO
void doScroll(ep, id)
EditorPtr ep; XtIntervalId *id;
#else
void doScroll(EditorPtr ep, XtIntervalId *id)
#endif
{
DeclareArgs(2);
Widget vbar, hbar;
int vbarValue, hbarValue, vbarMax, hbarMax;
int hbarSlider, hbarIncrement, hbarPage;
int vbarSlider, vbarIncrement, vbarPage;

    if (ep->scrollDirection && ep->workObscured)
    {
	XtAppAddTimeOut(XtWidgetToApplicationContext(ep->workarea), 
		SCROLL_INTERVAL, (XtTimerCallbackProc)doScroll, 
		(XtPointer)ep);
	ep->scrollInProgress = 1;
	FirstArg(XmNhorizontalScrollBar, &hbar);
	NextArg(XmNverticalScrollBar, &vbar);
	XtGetValues(ep->workScrollWin, Args, ArgCount);

	XmScrollBarGetValues(hbar, &hbarValue, &hbarSlider, &hbarIncrement,
		&hbarPage);
	XmScrollBarGetValues(vbar, &vbarValue, &vbarSlider, &vbarIncrement,
		&vbarPage);

	FirstArg(XmNmaximum, &vbarMax);
	XtGetValues(vbar, Args, ArgCount);
	FirstArg(XmNmaximum, &hbarMax);
	XtGetValues(hbar, Args, ArgCount);

	if (ep->scrollDirection & SCROLL_UP)
	    vbarValue -= ep->scale + 1;
	if (ep->scrollDirection & SCROLL_DOWN)
	    vbarValue += ep->scale + 1;
	if (ep->scrollDirection & SCROLL_LEFT)
	    hbarValue -= ep->scale + 1;
	if (ep->scrollDirection & SCROLL_RIGHT)
	    hbarValue += ep->scale + 1;

	if (vbarValue < 0)
	    vbarValue = 0;
	else if (vbarValue + vbarSlider > vbarMax)
	    vbarValue = vbarMax - vbarSlider;

	if (hbarValue < 0)
	    hbarValue = 0;
	else if (hbarValue + hbarSlider > hbarMax)
	    hbarValue = hbarMax - hbarSlider;

	XmScrollBarSetValues(hbar, hbarValue, hbarSlider, hbarIncrement,
		hbarPage, True);
	XmScrollBarSetValues(vbar, vbarValue, vbarSlider, vbarIncrement,
		vbarPage, True);
    }
    else
	ep->scrollInProgress = 0;
}

/*
** handle button press, release and motion events in the pixmap windows
*/
#ifdef _NO_PROTO
void inputImageEH(wid, ep, e)
Widget wid; EditorPtr ep; XEvent *e;
#else
void inputImageEH(Widget wid, EditorPtr ep, XEvent *e)
#endif
{
int x, y, cx, cy;
unsigned int state;
static int ignoreTillRelease;
int scroll = 0;

    if (e->type == ButtonPress || e->type == ButtonRelease)
    {
	if (wid == ep->workarea)
	{
	    x = (e->xbutton.x /= (ep->scale + 1));
	    y = (e->xbutton.y /= (ep->scale + 1));
	}
	else
	{
	    x = e->xbutton.x;
	    y = e->xbutton.y;
	}
	state = e->xbutton.state;
	ep->scrollDirection = 0;
    }
    else
    {
	if (ep->workObscured && wid == ep->workarea)
	{
	    cx = e->xmotion.x;
	    cy = e->xmotion.y;
	    if (cx + ep->workWinX < 0 && ep->workWinX < 0)
		scroll |= SCROLL_LEFT;
	    if (cy + ep->workWinY < 0 && ep->workWinY < 0)
		scroll |= SCROLL_UP;
	    if (cx + ep->workWinX >= ep->clipWinW && cx + ep->workWinX < ep->workWinW)
		scroll |= SCROLL_RIGHT;
	    if (cy + ep->workWinY >= ep->clipWinH && cy + ep->workWinY < ep->workWinH)
		scroll |= SCROLL_DOWN;
	    if (scroll)
	    {
		ep->scrollDirection = scroll;
		if (!ep->scrollInProgress)
		{
		    XtAppAddTimeOut(XtWidgetToApplicationContext(ep->workarea), 
			    SCROLL_INTERVAL, (XtTimerCallbackProc)doScroll,
			    (XtPointer)ep);
		    ep->scrollInProgress = 1;
		}
	    }
	    ep->scrollDirection = scroll;
	}
	if (wid == ep->workarea)
	{
	    x = (e->xmotion.x /= (ep->scale + 1));
	    y = (e->xmotion.y /= (ep->scale + 1));
	}
	else
	{
	    x = e->xmotion.x;
	    y = e->xmotion.y;
	}
	state = e->xmotion.state;
    }
    if (e->type == ButtonPress)
    {
	if (state & Mod1Mask)		/* use to set color */
	{
	    Pixel p;
	    ignoreTillRelease = True;
	    p = XGetPixel(ep->image, x, y);
	    setCurrentPixel((int)p);
	}
	else if (state & ControlMask)	/* use to move around image */
	{
	    ignoreTillRelease = True;
	}
	else
	    ignoreTillRelease = False;
    }
    if (!ep->buttonPressFunction)
	ignoreTillRelease = True;

    if (ignoreTillRelease)
	return;

    if (e->type == ButtonPress)
	(*(ep->buttonPressFunction))(ep, &e->xbutton);
    else if (e->type == ButtonRelease)
	(*(ep->buttonReleaseFunction))(ep, &e->xbutton);
    else if (state & (Button1Mask | Button2Mask | Button3Mask))
	(*(ep->buttonMotionFunction))(ep, &e->xmotion);
}

/*
** set a given pixel to a given value
*/
#ifdef _NO_PROTO
void setPixel(ep, pix, x, y)
EditorPtr ep; int pix; int x; int y;
#else
void setPixel(EditorPtr ep, int pix, int x, int y)
#endif
{
    /* clip to pixmap */
    if (x < 0 || y < 0 || x >= ep->pw || y >= ep->ph)
	return;
    XPutPixel(ep->image, x, y, pix);
    XSetForeground(display, ep->gc, pix);
    XDrawPoint(display, XtWindow(ep->pixarea), ep->gc, x, y);
    XDrawPoint(display, ep->pixmap, ep->gc, x, y);
    XFillRectangle(display, XtWindow(ep->workarea), ep->gc, 
	1 + x*(ep->scale+1), 1 + y*(ep->scale+1), ep->scale, ep->scale);
}


/*
** blit a image to the screen
** all non-zero pixels are blitted
*/
#ifdef _NO_PROTO
void blitBitmapImage(ep, srcImage, toPixmap, dx, dy, xmax, ymax)
EditorPtr ep; XImage *srcImage; int toPixmap; int dx; int dy; int xmax; int ymax;
#else
void blitBitmapImage(EditorPtr ep, XImage *srcImage, int toPixmap, int dx, int dy, int xmax, int ymax)
#endif
{
int x, y, ymin;
unsigned long p;
XPoint pts[MAX_PTS];
XRectangle rects[MAX_PTS];
int np = 0;

    XSetForeground(display, ep->gc, ep->fg);

    /* clip intersection of bitmap and pixmap */
    x = (dx < 0) ? -dx : 0;
    y = ymin = (dy < 0) ? -dy : 0;

    if (xmax + dx > ep->pw)
	xmax = (ep->pw - dx);
    if (ymax + dy > ep->ph)
	ymax = (ep->ph - dy);

    for (; x<xmax; x++)
    {
	for (y=ymin; y<ymax; y++)
	{
	    p = XGetPixel(srcImage, x, y);
	    if (p == 1)
	    {
		pts[np].x = x + dx;
		pts[np].y = y + dy;
		rects[np].x = 1 + (x + dx)*(ep->scale+1);
		rects[np].y = 1 + (y + dy)*(ep->scale+1);
		rects[np].width = rects[np].height = ep->scale;
		if (++np == MAX_PTS)
		{
		    XDrawPoints(display, XtWindow(ep->pixarea), ep->gc, 
				pts, np, CoordModeOrigin);
		    XFillRectangles(display, XtWindow(ep->workarea), ep->gc, 
				rects, np);
		    if (toPixmap)
			XDrawPoints(display, ep->pixmap, ep->gc, 
				pts, np, CoordModeOrigin);
		    np = 0;
		}
	    }
	}
    }
    if (np)
    {
	XDrawPoints(display, XtWindow(ep->pixarea), ep->gc, pts, np, CoordModeOrigin);
	XFillRectangles(display, XtWindow(ep->workarea), ep->gc, rects, np);
	if (toPixmap)
	    XDrawPoints(display, ep->pixmap, ep->gc, pts, np, CoordModeOrigin);
    }
    if (toPixmap)
	XGetSubImage(display, ep->pixmap, 0, 0, ep->pw, ep->ph, 
	    AllPlanes, ZPixmap, ep->image, 0, 0);
}

/*
** restore the original data based on what pixels were
** set in the srcImage
*/
#ifdef _NO_PROTO
void unblitBitmapImage(ep, srcImage, dx, dy, xmax, ymax)
EditorPtr ep; XImage *srcImage; int dx; int dy; int xmax; int ymax;
#else
void unblitBitmapImage(EditorPtr ep, XImage *srcImage, int dx, int dy, int xmax, int ymax)
#endif
{
int x, y, ymin;
unsigned long p, lastp = 0xffff;
XRectangle rects[MAX_PTS];
int nr = 0;

    /* clip area */
    x = (dx < 0) ? -dx : 0;
    y = ymin = (dy < 0) ? -dy : 0;

    if (xmax + dx > ep->pw)
	xmax = (ep->pw - dx);
    if (ymax + dy > ep->ph)
	ymax = (ep->ph - dy);

    for (; x<xmax; x++)
    {
	for (y = ymin; y<ymax; y++)
	{
	    p = XGetPixel(srcImage, x, y);
	    if (p)
	    {
		p = XGetPixel(ep->image, x + dx, y + dy);
		if (lastp != p)
		{
		    if (nr)
		    {
			XFillRectangles(display, XtWindow(ep->workarea), ep->gc, 
			    rects, nr);
			nr = 0;
		    }
		    XSetForeground(display, ep->gc, lastp = p);
		}
		rects[nr].x = 1 + (x + dx) * (ep->scale + 1);
		rects[nr].y = 1 + (y + dy) * (ep->scale + 1);
		rects[nr].width = rects[nr].height = ep->scale;
		if (++nr == MAX_PTS)
		{
		    XFillRectangles(display, XtWindow(ep->workarea), ep->gc, 
			rects, nr);
		    nr = 0;
		}
	    }
	}
    }
    if (nr)
	XFillRectangles(display, XtWindow(ep->workarea), ep->gc, rects, nr);
    XCopyArea(display, ep->pixmap, XtWindow(ep->pixarea), ep->gc, 0, 0, ep->pw, ep->ph, 0, 0);
}

/*
** blit the work pixmap to the screen
** only pixels set to 1 are copied
*/
#ifdef _NO_PROTO
void blitWorkPixmap(ep, toPixmap, alreadyInImage, srcx, srcy, srcw, srch)
EditorPtr ep; int toPixmap; int alreadyInImage;
	int srcx; int srcy; int srcw; int srch;
#else
void blitWorkPixmap(EditorPtr ep, int toPixmap, int alreadyInImage,
	int srcx, int srcy, int srcw, int srch)
#endif
{
int x, y;
unsigned long p;
XPoint pts[MAX_PTS];
XRectangle rects[MAX_PTS];
int np = 0;

    if (!alreadyInImage)
	XGetSubImage(display, ep->workpixmap, srcx, srcy, srcw, srch,
	    AllPlanes, ZPixmap, ep->workimage, srcx, srcy);

    XSetForeground(display, ep->gc, ep->fg);

    for (x=srcx; x<srcx + srcw; x++)
    {
	for (y=srcy; y<srcy + srch; y++)
	{
	    p = XGetPixel(ep->workimage, x, y);
	    if (p == 1)
	    {
		pts[np].x = x;
		pts[np].y = y;
		rects[np].x = 1 + x*(ep->scale+1);
		rects[np].y = 1 + y*(ep->scale+1);
		rects[np].width = rects[np].height = ep->scale;
		if (++np == MAX_PTS)
		{
		    XDrawPoints(display, XtWindow(ep->pixarea), ep->gc, 
				pts, np, CoordModeOrigin);
		    XFillRectangles(display, XtWindow(ep->workarea), ep->gc, 
				rects, np);
		    if (toPixmap)
			XDrawPoints(display, ep->pixmap, ep->gc, 
				pts, np, CoordModeOrigin);
		    np = 0;
		}
	    }
	}
    }
    if (np)
    {
	XDrawPoints(display, XtWindow(ep->pixarea), ep->gc, pts, np, CoordModeOrigin);
	XFillRectangles(display, XtWindow(ep->workarea), ep->gc, rects, np);
	if (toPixmap)
	    XDrawPoints(display, ep->pixmap, ep->gc, pts, np, CoordModeOrigin);
    }
    if (toPixmap)
	XGetSubImage(display, ep->pixmap, srcx, srcy, srcw, srch,
	    AllPlanes, ZPixmap, ep->image, srcx, srcy);
}

/*
** restore the original data based on what pixels were
** changed in the work pixmap
*/
#ifdef _NO_PROTO
void unblitWorkPixmap(ep, srcx, srcy, srcw, srch)
EditorPtr ep; int srcx; int srcy; int srcw; int srch;
#else
void unblitWorkPixmap(EditorPtr ep,
	int srcx, int srcy, int srcw, int srch)
#endif
{
int x, y;
unsigned long p, lastp = 0xffff;
XRectangle rects[MAX_PTS];
int nr = 0;

    for (x=srcx; x<srcx + srcw; x++)
    {
	for (y=srcy; y<srcy + srch; y++)
	{
	    p = XGetPixel(ep->workimage, x, y);
	    if (p == 1)
	    {
		p = XGetPixel(ep->image, x, y);
		if (lastp != p)
		{
		    if (nr)
		    {
			XFillRectangles(display, XtWindow(ep->workarea), ep->gc, 
			    rects, nr);
			nr = 0;
		    }
		    XSetForeground(display, ep->gc, lastp = p);
		}
		rects[nr].x = 1 + x * (ep->scale + 1);
		rects[nr].y = 1 + y * (ep->scale + 1);
		rects[nr].width = rects[nr].height = ep->scale;
		if (++nr == MAX_PTS)
		{
		    XFillRectangles(display, XtWindow(ep->workarea), ep->gc, 
			rects, nr);
		    nr = 0;
		}
	    }
	}
    }
    if (nr)
	XFillRectangles(display, XtWindow(ep->workarea), ep->gc, rects, nr);
    XCopyArea(display, ep->pixmap, XtWindow(ep->pixarea), ep->gc, srcx, srcy, srcw, srch, srcx, srcy);
}

/*
** place image onto screen
*/
#ifdef _NO_PROTO
void blitImage(ep, dataimage, dx, dy, opaq, toPixmap)
EditorPtr ep; XImage *dataimage; int dx; int dy; int opaq; int toPixmap;
#else
void blitImage(EditorPtr ep, XImage *dataimage, int dx, int dy, 
	int opaq, int toPixmap)
#endif
{
int x, y, ymin, xmax, ymax;
unsigned long p, lastp = 0xffff;
XPoint pts[MAX_PTS];
XRectangle rects[MAX_PTS];
int np = 0;

    /* clip area */
    x = (dx < 0) ? -dx : 0;
    y = ymin = (dy < 0) ? -dy : 0;
    xmax = dataimage->width;
    ymax = dataimage->height;

    if (xmax + dx > ep->pw)
	xmax = (ep->pw - dx);
    if (ymax + dy > ep->ph)
	ymax = (ep->ph - dy);

    for (; x<xmax; x++)
    {
	for (y = ymin; y<ymax; y++)
	{
	    p = XGetPixel(dataimage, x, y);
	    if (!opaq && !p)
		continue;
	    if (lastp != p)
	    {
		if (np)
		{
		    XDrawPoints(display, XtWindow(ep->pixarea), ep->gc, 
			    pts, np, CoordModeOrigin);
		    XFillRectangles(display, XtWindow(ep->workarea), ep->gc, 
			rects, np);
		    if (toPixmap)
			XDrawPoints(display, ep->pixmap, ep->gc, 
				pts, np, CoordModeOrigin);
		    np = 0;
		}
		XSetForeground(display, ep->gc, lastp = p);
	    }
	    pts[np].x = x + dx;
	    pts[np].y = y + dy;
	    rects[np].x = 1 + (x + dx) * (ep->scale + 1);
	    rects[np].y = 1 + (y + dy) * (ep->scale + 1);
	    rects[np].width = rects[np].height = ep->scale;
	    if (++np == MAX_PTS)
	    {
		XDrawPoints(display, XtWindow(ep->pixarea), ep->gc, 
			pts, np, CoordModeOrigin);
		XFillRectangles(display, XtWindow(ep->workarea), ep->gc, 
		    rects, np);
		np = 0;
		if (toPixmap)
		    XDrawPoints(display, ep->pixmap, ep->gc, 
			    pts, np, CoordModeOrigin);
	    }
	}
    }
    if (np)
    {
	XDrawPoints(display, XtWindow(ep->pixarea), ep->gc, pts, np, CoordModeOrigin);
	XFillRectangles(display, XtWindow(ep->workarea), ep->gc, rects, np);
	if (toPixmap)
	    XDrawPoints(display, ep->pixmap, ep->gc, pts, np, CoordModeOrigin);
    }
    if (toPixmap)
	XGetSubImage(display, ep->pixmap, 0, 0, ep->pw, ep->ph, 
	    AllPlanes, ZPixmap, ep->image, 0, 0);
}

/*
** restore are where image was drawn
*/
#ifdef _NO_PROTO
void unblitImage(ep, dataimage, dx, dy, opaq)
EditorPtr ep; XImage *dataimage; int dx; int dy; int opaq;
#else
void unblitImage(EditorPtr ep, XImage *dataimage, int dx, int dy, int opaq)
#endif
{
int x, y, ymin, xmax, ymax;
unsigned long p, lastp = 0xffff;
XRectangle rects[MAX_PTS];
int nr = 0;

    /* clip area */
    x = (dx < 0) ? -dx : 0;
    y = ymin = (dy < 0) ? -dy : 0;
    xmax = dataimage->width;
    ymax = dataimage->height;

    if (xmax + dx > ep->pw)
	xmax = (ep->pw - dx);
    if (ymax + dy > ep->ph)
	ymax = (ep->ph - dy);

    for (; x<xmax; x++)
    {
	for (y = ymin; y<ymax; y++)
	{
	    p = XGetPixel(dataimage, x, y);
	    if (!opaq && !p)
		continue;
	    p = XGetPixel(ep->image, x + dx, y + dy);
	    if (lastp != p)
	    {
		if (nr)
		{
		    XFillRectangles(display, XtWindow(ep->workarea), ep->gc, 
			rects, nr);
		    nr = 0;
		}
		XSetForeground(display, ep->gc, lastp = p);
	    }
	    rects[nr].x = 1 + (x + dx) * (ep->scale + 1);
	    rects[nr].y = 1 + (y + dy) * (ep->scale + 1);
	    rects[nr].width = rects[nr].height = ep->scale;
	    if (++nr == MAX_PTS)
	    {
		XFillRectangles(display, XtWindow(ep->workarea), ep->gc, 
		    rects, nr);
		nr = 0;
	    }
	}
    }
    if (nr)
	XFillRectangles(display, XtWindow(ep->workarea), ep->gc, rects, nr);
    XCopyArea(display, ep->pixmap, XtWindow(ep->pixarea), ep->gc, 0, 0, ep->pw, ep->ph, 0, 0);
}

/*
** clear the work pixmap to all zeros
*/
#ifdef _NO_PROTO
void clearWorkPixmap(ep)
EditorPtr ep;
#else
void clearWorkPixmap(EditorPtr ep)
#endif
{
    XSetForeground(display, ep->workgc, 0);
    XFillRectangle(display, ep->workpixmap, ep->workgc, 0, 0, ep->pw, ep->ph);
    XSetForeground(display, ep->workgc, 1);
}

/*
** clear the work pixmap to all zeros
*/
#ifdef _NO_PROTO
void clearWorkImage(ep)
EditorPtr ep;
#else
void clearWorkImage(EditorPtr ep)
#endif
{
int x, y;
    for (x=0; x<ep->pw; x++)
	for (y=0; y<ep->ph; y++)
	    XPutPixel(ep->workimage, x, y, 0);
}

/********************************************************************
** drawPrim class functions
********************************************************************/
#ifdef _NO_PROTO
void drawPrimBPF(ep, e)
EditorPtr ep; XButtonEvent *e;
#else
void drawPrimBPF(EditorPtr ep, XButtonEvent *e)
#endif
{
register int x, y;
    saveState(ep);
    x = e->x;
    y = e->y;
    if (e->button == 2)
    {
	Pixel p;
	ep->fg = -1;
	p = XGetPixel(ep->image, x, y);
	setCurrentPixel((int)p);
    }
    else
    {
	ep->fg = (int)(e->button == 1 ? curPixel : 0);
	ep->lastx = x;
	ep->lasty = y;
	setPixel(ep, ep->fg, x, y);
    }
}

/* ARGSUSED */
#ifdef _NO_PROTO
void drawPrimBRF(ep, e)
EditorPtr ep; XButtonEvent *e;
#else
void drawPrimBRF(EditorPtr ep, XButtonEvent *e)
#endif
{
    return;
}

#ifdef _NO_PROTO
void drawPrimBMF(ep, e)
EditorPtr ep; XMotionEvent *e;
#else
void drawPrimBMF(EditorPtr ep, XMotionEvent *e)
#endif
{
    if (ep->fg != -1 && (e->x != ep->lastx || e->y != ep->lasty))
	setPixel(ep, ep->fg, e->x, e->y);
}

/********************************************************************
** linePrim class functions
********************************************************************/
#ifdef _NO_PROTO
void drawLine(im, pw, ph, x1, y1, x2, y2)
XImage *im; int pw; int ph; int x1; int y1; int x2; int y2;
#else
void drawLine(XImage *im, int pw, int ph, int x1, int y1, int x2, int y2)
#endif
{
int x, y;
int dx, dy;
double f, df;
    /* horizontal line? */
    if (x1 == x2)
    {
	if (y2 < y1)
	{
	    y = y1; y1 = y2; y2 = y;
	}
	if (y1 < 0)
	    y1 = 0;
	if (y2 >= ph)
	    y2 = ph-1;
	for (y=y1; y<=y2; y++)
	    XPutPixel(im, x1, y, 1);
	return;
    }
    /* vertical line? */
    else if (y1 == y2)
    {
	if (x2 < x1)
	{
	    x = x1; x1 = x2; x2 = x;
	}
	if (x1 < 0)
	    x1 = 0;
	if (x2 >= pw)
	    x2 = pw-1;
	for (x=x1; x<=x2; x++)
	    XPutPixel(im, x, y1, 1);
	return;
    }

    dx = (x1 < x2) ? (x2 - x1) : (x1 - x2);
    dy = (y1 < y2) ? (y2 - y1) : (y1 - y2);

    /* increment by x=1? */
    if (dx >= dy)
    {
	if (x2 < x1)
	{
	    x = x1; x1 = x2; x2 = x;
	    y = y1; y1 = y2; y2 = y;
	}
	df = (double)(y2 - y1) / (double)(x2 - x1);
	f = y1 + 0.5;
	for (x=x1; x<=x2; x++)
	{
	    if (x >=0 && x < pw)
	    {
		if (f >= 0.0 && f < ph)
		{
		    y = (int)f;
		    XPutPixel(im, x, y, 1);
		}
	    }
	    f += df;
	}
	return;
    }
    else /* increment by y=1? */
    {
	if (y2 < y1)
	{
	    x = x1; x1 = x2; x2 = x;
	    y = y1; y1 = y2; y2 = y;
	}
	df = (double)(x2 - x1) / (double)(y2 - y1);
	f = x1 + 0.5;
	for (y=y1; y<=y2; y++)
	{
	    if (y >=0 && y < ph)
	    {
		if (f >= 0.0 && f < pw)
		{
		    x = (int)f;
		    XPutPixel(im, x, y, 1);
		}
	    }
	    f += df;
	}
	return;
    }
}

#ifdef _NO_PROTO
void linePrimBPF(ep, e)
EditorPtr ep; XButtonEvent *e;
#else
void linePrimBPF(EditorPtr ep, XButtonEvent *e)
#endif
{
int simpleLine = True;
    saveState(ep);
    if (e->button == 1)
	ep->fg = (int)curPixel;
    else if (e->button == 2)
    {
	ep->fg = (int)XGetPixel(ep->image, e->x, e->y);
	setCurrentPixel(ep->fg);
    }
    else
	ep->fg = 0;

    ep->lastx = ep->basex = e->x;
    ep->lasty = ep->basey = e->y;
    if (simpleLine)
    {
	clearWorkImage(ep);
	drawLine(ep->workimage, ep->pw, ep->ph, e->x, e->y, e->x, e->y);
	blitWorkPixmap(ep, False, True, 0, 0, ep->pw, ep->ph);
    }
    else
    {
	clearWorkPixmap(ep);
	XSetForeground(display, ep->workgc, 1);
	XDrawLine(display, ep->workpixmap, ep->workgc, e->x, e->y, e->x, e->y);
	blitWorkPixmap(ep, False, False, 0, 0, ep->pw, ep->ph);
    }
}

/* ARGSUSED */
#ifdef _NO_PROTO
void linePrimBRF(ep, e)
EditorPtr ep; XButtonEvent *e;
#else
void linePrimBRF(EditorPtr ep, XButtonEvent *e)
#endif
{
int simpleLine = True;
    if (simpleLine)
	blitWorkPixmap(ep, True, True, 0, 0, ep->pw, ep->ph);
    else
	blitWorkPixmap(ep, True, False, 0, 0, ep->pw, ep->ph);
}

#ifdef _NO_PROTO
void linePrimBMF(ep, e)
EditorPtr ep; XMotionEvent *e;
#else
void linePrimBMF(EditorPtr ep, XMotionEvent *e)
#endif
{
int simpleLine = True;
    if (e->x == ep->lastx && e->y == ep->lasty)
	return;
    unblitWorkPixmap(ep, 0, 0, ep->pw, ep->ph);

    ep->lastx = e->x;
    ep->lasty = e->y;
    if (simpleLine)
    {
	clearWorkImage(ep);
	drawLine(ep->workimage, ep->pw, ep->ph, 
		ep->basex, ep->basey, e->x, e->y);
	blitWorkPixmap(ep, False, True, 0, 0, ep->pw, ep->ph);
    }
    else
    {
	clearWorkPixmap(ep);
	XDrawLine(display, ep->workpixmap, ep->workgc, 
	    ep->basex, ep->basey, e->x, e->y);
	blitWorkPixmap(ep, False, False, 0, 0, ep->pw, ep->ph);
    }
}


/********************************************************************
** rectPrim class functions
********************************************************************/
#ifdef _NO_PROTO
void drawRect(im, pw, ph, x1, y1, w, h)
XImage *im; int pw; int ph; int x1; int y1; int w; int h;
#else
void drawRect(XImage *im, int pw, int ph, int x1, int y1, int w, int h)
#endif
{
int x, y;
int x2, y2;

    x2 = x1 + w;
    y2 = y1 + h;

    for (x=x1; x <= x2; x++)
    {
	if (x < 0)
	    continue;
	else if (x >= pw)
	    break;
	if (y1 >= 0)
	    XPutPixel(im, x, y1, 1);
	if (y2 < ph)
	    XPutPixel(im, x, y2, 1);
    }
    for (y=y1; y <= y2; y++)
    {
	if (y < 0)
	    continue;
	else if (y >= ph)
	    break;
	if (x1 >= 0)
	    XPutPixel(im, x1, y, 1);
	if (x2 < pw)
	    XPutPixel(im, x2, y, 1);
    }
}

#ifdef _NO_PROTO
void rectPrimBPF(ep, e)
EditorPtr ep; XButtonEvent *e;
#else
void rectPrimBPF(EditorPtr ep, XButtonEvent *e)
#endif
{
int simpleRect = True;
    saveState(ep);
    if (e->button == 1)
	ep->fg = (int)curPixel;
    else if (e->button == 2)
    {
	ep->fg = (int)XGetPixel(ep->image, e->x, e->y);
	setCurrentPixel(ep->fg);
    }
    else
	ep->fg = 0;

    ep->lastx = ep->basex = e->x;
    ep->lasty = ep->basey = e->y;
    if (simpleRect)
    {
	clearWorkImage(ep);
	drawRect(ep->workimage, ep->pw, ep->ph, e->x, e->y, 0, 0);
	blitWorkPixmap(ep, False, True, 0, 0, ep->pw, ep->ph);
    }
    else
    {
	clearWorkPixmap(ep);
	XSetForeground(display, ep->workgc, 1);
	XDrawRectangle(display, ep->workpixmap, ep->workgc, 
		    e->x, e->y, 0, 0);
	blitWorkPixmap(ep, False, False, 0, 0, ep->pw, ep->ph);
    }
}

/* ARGSUSED */
#ifdef _NO_PROTO
void rectPrimBRF(ep, e)
EditorPtr ep; XButtonEvent *e;
#else
void rectPrimBRF(EditorPtr ep, XButtonEvent *e)
#endif
{
int simpleRect = True;
    if (simpleRect)
	blitWorkPixmap(ep, True, True, 0, 0, ep->pw, ep->ph);
    else
	blitWorkPixmap(ep, True, False, 0, 0, ep->pw, ep->ph);
}

#ifdef _NO_PROTO
void rectPrimBMF(ep, e)
EditorPtr ep; XMotionEvent *e;
#else
void rectPrimBMF(EditorPtr ep, XMotionEvent *e)
#endif
{
int x, y, w, h;
int simpleRect = True;
    if (e->x == ep->lastx && e->y == ep->lasty)
	return;
    unblitWorkPixmap(ep, 0, 0, ep->pw, ep->ph);

    if (e->x < ep->basex)
    {
	x = e->x;
	w = ep->basex - x;
    }
    else
    {
	x = ep->basex;
	w = e->x - x;
    }
    if (e->y < ep->basey)
    {
	y = e->y;
	h = ep->basey - y;
    }
    else
    {
	y = ep->basey;
	h = e->y - y;
    }
    ep->lastx = e->x;
    ep->lasty = e->y;
    if (simpleRect)
    {
	clearWorkImage(ep);
	drawRect(ep->workimage, ep->pw, ep->ph, x, y, w, h);
	blitWorkPixmap(ep, False, True, 0, 0, ep->pw, ep->ph);
    }
    else
    {
	clearWorkPixmap(ep);
	XDrawRectangle(display, ep->workpixmap, ep->workgc, x, y, w, h);
	blitWorkPixmap(ep, False, False, 0, 0, ep->pw, ep->ph);
    }
}

/********************************************************************
** fillRectPrim class functions
********************************************************************/
#ifdef _NO_PROTO
void fillRect(im, pw, ph, x1, y1, w, h)
XImage *im; int pw; int ph; int x1; int y1; int w; int h;
#else
void fillRect(XImage *im, int pw, int ph, int x1, int y1, int w, int h)
#endif
{
int x, y;
int x2, y2;
    x2 = x1 + w;
    y2 = y1 + h;
    if (x1 < 0)
	x1 = 0;
    if (y1 < 0)
	y1 = 0;
    if (x2 > pw)
	x2 = pw;
    if (y2 > ph)
	y2 = ph;
	
    for (x=x1; x<x2; x++)
	for (y=y1; y<y2; y++)
	    XPutPixel(im, x, y, 1);
}

#ifdef _NO_PROTO
void fillRectPrimBPF(ep, e)
EditorPtr ep; XButtonEvent *e;
#else
void fillRectPrimBPF(EditorPtr ep, XButtonEvent *e)
#endif
{
int simpleRect = True;
    saveState(ep);
    if (e->button == 1)
	ep->fg = (int)curPixel;
    else if (e->button == 2)
    {
	ep->fg = (int)XGetPixel(ep->image, e->x, e->y);
	setCurrentPixel(ep->fg);
    }
    else
	ep->fg = 0;

    ep->lastx = ep->basex = e->x;
    ep->lasty = ep->basey = e->y;
    if (simpleRect)
    {
	clearWorkImage(ep);
	fillRect(ep->workimage, ep->pw, ep->ph, e->x, e->y, 1, 1);
	blitWorkPixmap(ep, False, True, 0, 0, ep->pw, ep->ph);
    }
    else
    {
	clearWorkPixmap(ep);
	XSetForeground(display, ep->workgc, 1);
	XFillRectangle(display, ep->workpixmap, ep->workgc, e->x, e->y, 1, 1);
	blitWorkPixmap(ep, False, False, 0, 0, ep->pw, ep->ph);
    }
}

/* ARGSUSED */
#ifdef _NO_PROTO
void fillRectPrimBRF(ep, e)
EditorPtr ep; XButtonEvent *e;
#else
void fillRectPrimBRF(EditorPtr ep, XButtonEvent *e)
#endif
{
int simpleRect = True;
    if (simpleRect)
	blitWorkPixmap(ep, True, True, 0, 0, ep->pw, ep->ph);
    else
	blitWorkPixmap(ep, True, False, 0, 0, ep->pw, ep->ph);
}

#ifdef _NO_PROTO
void fillRectPrimBMF(ep, e)
EditorPtr ep; XMotionEvent *e;
#else
void fillRectPrimBMF(EditorPtr ep, XMotionEvent *e)
#endif
{
int x, y, w, h;
int simpleRect = True;
    if (e->x == ep->lastx && e->y == ep->lasty)
	return;
    unblitWorkPixmap(ep, 0, 0, ep->pw, ep->ph);

    if (e->x < ep->basex)
    {
	x = e->x;
	w = ep->basex - x + 1;
    }
    else
    {
	x = ep->basex;
	w = e->x - x + 1;
    }
    if (e->y < ep->basey)
    {
	y = e->y;
	h = ep->basey - y + 1;
    }
    else
    {
	y = ep->basey;
	h = e->y - y + 1;
    }
    ep->lastx = e->x;
    ep->lasty = e->y;
    if (simpleRect)
    {
	clearWorkImage(ep);
	fillRect(ep->workimage, ep->pw, ep->ph, x, y, w, h);
	blitWorkPixmap(ep, False, True, 0, 0, ep->pw, ep->ph);
    }
    else
    {
	clearWorkPixmap(ep);
	XFillRectangle(display, ep->workpixmap, ep->workgc, x, y, w, h);
	blitWorkPixmap(ep, False, False, 0, 0, ep->pw, ep->ph);
    }
}


/********************************************************************
** floodPrim class functions
********************************************************************/
static XImage *floodImage;

#ifdef _NO_PROTO
Pixel pixelread(x, y)
int x; int y;
#else
Pixel pixelread(int x, int y)
#endif
{
    return(XGetPixel(floodImage, x, y));
}

#ifdef _NO_PROTO
void pixelwrite(x, y, p)
int x; int y; Pixel p;
#else
void pixelwrite(int x, int y, Pixel p)
#endif
{
    XPutPixel(floodImage, x, y, p);
}

/*
 * A Seed Fill Algorithm
 * by Paul Heckbert
 * from "Graphics Gems", Academic Press, 1990
 */

typedef struct {		/* window: a discrete 2-D rectangle */
    int x0, y0;			/* xmin and ymin */
    int x1, y1;			/* xmax and ymax (inclusive) */
} FillWindow;


typedef struct {short y, xl, xr, dy;} Segment;
/*
 * Filled horizontal segment of scanline y for xl<=x<=xr.
 * Parent segment was on line y-dy.  dy=1 or -1
 */

#define MAX_DEPTH 1000		/* max depth of stack */

#define PUSH(Y, XL, XR, DY)	/* push new segment on stack */ \
    if (sp<stack+MAX_DEPTH && Y+(DY)>=win->y0 && Y+(DY)<=win->y1) \
    {sp->y = Y; sp->xl = XL; sp->xr = XR; sp->dy = DY; sp++;}

#define POP(Y, XL, XR, DY)	/* pop segment off stack */ \
    {sp--; Y = sp->y+(DY = sp->dy); XL = sp->xl; XR = sp->xr;}

/*
 * fill: set the pixel at (x,y) and all of its 4-connected neighbors
 * with the same pixel value to the new pixel value nv.
 * A 4-connected neighbor is a pixel above, below, left, or right of a pixel.
 */

#ifdef _NO_PROTO
void fill(x, y, win, nv)
int x; int y; FillWindow *win; Pixel nv;
#else
void fill(int x, int y, FillWindow *win, Pixel nv)
#endif
{
    int l, x1, x2, dy;
    Pixel ov;	/* old pixel value */
    Segment stack[MAX_DEPTH], *sp = stack;	/* stack of filled segments */

    ov = pixelread(x, y);		/* read pv at seed point */
    if (ov==nv || x<win->x0 || x>win->x1 || y<win->y0 || y>win->y1) return;
    PUSH(y, x, x, 1);			/* needed in some cases */
    PUSH(y+1, x, x, -1);		/* seed segment (popped 1st) */

    while (sp>stack) {
	/* pop segment off stack and fill a neighboring scan line */
	POP(y, x1, x2, dy);
	/*
	 * segment of scan line y-dy for x1<=x<=x2 was previously filled,
	 * now explore adjacent pixels in scan line y
	 */
	for (x=x1; x>=win->x0 && pixelread(x, y)==ov; x--)
	    pixelwrite(x, y, nv);
	if (x>=x1) goto skip;
	l = x+1;
	if (l<x1) PUSH(y, l, x1-1, -dy);		/* leak on left? */
	x = x1+1;
	do {
	    for (; x<=win->x1 && pixelread(x, y)==ov; x++)
		pixelwrite(x, y, nv);
	    PUSH(y, l, x-1, dy);
	    if (x>x2+1) PUSH(y, x2+1, x-1, -dy);	/* leak on right? */
skip:	    for (x++; x<=x2 && pixelread(x, y)!=ov; x++);
	    l = x;
	} while (x<=x2);
    }
}

#ifdef _NO_PROTO
void flattenImage(im, w, h, zeroPix, onePix, twoPix, defValue)
XImage *im; int w; int h; 
	int zeroPix; int onePix; int twoPix; int defValue;
#else
void flattenImage(XImage *im, int w, int h, 
	int zeroPix, int onePix, int twoPix, int defValue)
#endif
{
int x, y, p;
    for (x=0; x<w; x++)
	for (y=0; y<h; y++)
	{
	    p = (int)XGetPixel(im, x, y);
	    if (p == zeroPix)
		XPutPixel(im, x, y, 0);
	    else if (p == onePix)
		XPutPixel(im, x, y, 1);
	    else if (p == twoPix)
		XPutPixel(im, x, y, 2);
	    else
		XPutPixel(im, x, y, defValue);
	}
}

#ifdef _NO_PROTO
void floodPrimIPF(ep)
EditorPtr ep;
#else
void floodPrimIPF(EditorPtr ep)
#endif
{
    sayMessage(ep, "Button 2 seeds boundary fill");
    ep->basex = ep->basey = 0;
}

#ifdef _NO_PROTO
void floodPrimBPF(ep, e)
EditorPtr ep; XButtonEvent *e;
#else
void floodPrimBPF(EditorPtr ep, XButtonEvent *e)
#endif
{
    FillWindow fillwin;

    if (e->button == 1)
	ep->fg = (int)curPixel;
    else if (e->button == 2)
    {
	ep->basex = e->x;
	ep->basey = e->y;
	return;
    }
    else
	ep->fg = 0;

    saveState(ep);
    XGetSubImage(display, ep->pixmap, 0, 0, ep->pw, ep->ph, 
	AllPlanes, ZPixmap, ep->workimage, 0, 0);

    fillwin.x0 = fillwin.y0 = 0;
    fillwin.x1 = ep->pw-1;
    fillwin.y1 = ep->ph-1;
    floodImage = ep->workimage;
    /* convert image to binary  */
    if (ep->fillToBoundary)
    {
	Pixel p;
	p = XGetPixel(floodImage, e->x, e->y);
	flattenImage(floodImage, ep->pw, ep->ph, -1, -1, (int)p, 0);
	fill(ep->basex, ep->basey, &fillwin, 1);
    }
    else
    {
	Pixel p;
	p = XGetPixel(floodImage, e->x, e->y);
	flattenImage(floodImage, ep->pw, ep->ph, (int)p, -1, -1, 2);
	fill(e->x, e->y, &fillwin, 1);
    }
    blitWorkPixmap(ep, True, True, 0, 0, ep->pw, ep->ph);
}

/* ARGSUSED */
#ifdef _NO_PROTO
void floodPrimBRF(ep, e)
EditorPtr ep; XButtonEvent *e;
#else
void floodPrimBRF(EditorPtr ep, XButtonEvent *e)
#endif
{
    return;
}

/* ARGSUSED */
#ifdef _NO_PROTO
void floodPrimBMF(ep, e)
EditorPtr ep; XMotionEvent *e;
#else
void floodPrimBMF(EditorPtr ep, XMotionEvent *e)
#endif
{
    return;
}

/********************************************************************
** bitmapPrim class functions
********************************************************************/

#ifdef _NO_PROTO
void loadBitmap(ep, filename)
EditorPtr ep; String filename;
#else
void loadBitmap(EditorPtr ep, String filename)
#endif
{
unsigned int w, h;
int dummy, status;
Pixmap bitmap;

    status = XReadBitmapFile(display, XtWindow(ep->workarea), filename, 
		&w, &h, &bitmap, &dummy, &dummy);
    if (status != BitmapSuccess)
    {
	char buf[512];
	sprintf(buf, "Cannot Load Bitmap File\n'%s'", filename);
	popupMessage(XPE_Error, buf, NULL, NULL, NULL);
	return;
    }
    if (ep->bitmapimage)
	XDestroyImage(ep->bitmapimage);
    ep->bitmapimage = XGetImage(display, bitmap, 0, 0, w, h,
		    AllPlanes, ZPixmap);
    ep->bitw = w;
    ep->bith = h;
    XFreePixmap(display, bitmap);
}

/* ARGSUSED */
#ifdef _NO_PROTO
void loadBitmapCB(wid, unused, unused_too)
Widget wid; XtPointer unused; XtPointer unused_too;
#else
void loadBitmapCB(Widget wid, XtPointer unused, XtPointer unused_too)
#endif
{
EditorPtr ep;

    ep = getEditor(wid);
    askFile(loadBitmap, "Load Bitmap", ep, "*.xbm");
}

#ifdef _NO_PROTO
void bitmapPrimIPF(ep)
EditorPtr ep;
#else
void bitmapPrimIPF(EditorPtr ep)
#endif
{
    if (!ep->bitmapimage)
	sayMessage(ep, "No bitmap image loaded");
}

#ifdef _NO_PROTO
void bitmapPrimBPF(ep, e)
EditorPtr ep; XButtonEvent *e;
#else
void bitmapPrimBPF(EditorPtr ep, XButtonEvent *e)
#endif
{
    if (e->button == 1)
	ep->fg = (int)curPixel;
    else if (e->button == 2)
    {
	/* askForBitmap((void *)ep); */
	askFile(loadBitmap, "Load Bitmap", ep, "*.xbm");
	return;
    }
    else
	ep->fg = 0;
    if (!ep->bitmapimage)
    {
	popupMessage(XPE_Warning, "No Bitmap Loaded!", NULL, NULL, NULL);
	return;
    }
    saveState(ep);

    ep->lastx = e->x;
    ep->lasty = e->y;
    blitBitmapImage(ep, ep->bitmapimage, False, e->x, e->y, ep->bitw, ep->bith);
}

#ifdef _NO_PROTO
void bitmapPrimBRF(ep, e)
EditorPtr ep; XButtonEvent *e;
#else
void bitmapPrimBRF(EditorPtr ep, XButtonEvent *e)
#endif
{
    if (!ep->bitmapimage || e->button == 2)
	return;
    blitBitmapImage(ep, ep->bitmapimage, True, ep->lastx, ep->lasty, ep->bitw, ep->bith);
}

#ifdef _NO_PROTO
void bitmapPrimBMF(ep, e)
EditorPtr ep; XMotionEvent *e;
#else
void bitmapPrimBMF(EditorPtr ep, XMotionEvent *e)
#endif
{
    if (!ep->bitmapimage || e->state & Button2Mask)
	return;
    if (e->x == ep->lastx && e->y == ep->lasty)
	return;
    unblitBitmapImage(ep, ep->bitmapimage, ep->lastx, ep->lasty, ep->bitw, ep->bith);

    ep->lastx = e->x;
    ep->lasty = e->y;
    blitBitmapImage(ep, ep->bitmapimage, False, e->x, e->y, ep->bitw, ep->bith);
}

/********************************************************************
** textPrim class functions
********************************************************************/
#ifdef _NO_PROTO
void getTextEntry(ep, x, y)
EditorPtr ep; int x; int y;
#else
void getTextEntry(EditorPtr ep, int x, int y)
#endif
{
    DeclareArgs(2);
    Dimension w, h, tstw, tsth;

    x *= (ep->scale + 1);
    y *= (ep->scale + 1);

    FirstArg(XmNwidth, &w);
    NextArg(XmNheight, &h);
    XtGetValues(ep->textEntry, Args, ArgCount);

    /* account for scrolling */
    x += ep->workWinX;
    y += ep->workWinY;

    x -= (w/2);
    y -= (h/2);

    /* clip to the smaller */
    if (ep->workWinW > ep->clipWinW)
	tstw = ep->clipWinW;
    else
	tstw = ep->workWinW;
    if (ep->workWinH > ep->clipWinH)
	tsth = ep->clipWinH;
    else
	tsth = ep->workWinH;

    if (x + w > tstw)
	x = tstw - w;
    if (y + h > tsth)
	y = tsth - h;
    if (x < 0)
	x = 0;
    if (y < 0)
	y = 0;

    FirstArg(XmNx, x - ep->workWinX);
    NextArg(XmNy, y - ep->workWinY);
    XtSetValues(ep->textEntry, Args, ArgCount);

    XtManageChild(ep->textEntry);
    XtAddGrab(ep->textEntry, True, False);
    XmProcessTraversal(ep->textEntry, XmTRAVERSE_CURRENT);
}

#ifdef _NO_PROTO
void textPrimIPF(ep)
EditorPtr ep;
#else
void textPrimIPF(EditorPtr ep)
#endif
{
    char msg[1024];
    if (!ep->enteredText)
	sayMessage(ep, "No text selected - Use Btn 2");
    else
    {
	sprintf(msg, "Text: '%s'", ep->enteredText);
	sayMessage(ep, msg);
    }
}

/* ARGSUSED */
#ifdef _NO_PROTO
void textDoneCB(wid, unused, unused_too)
Widget wid; XtPointer unused; XtPointer unused_too;
#else
void textDoneCB(Widget wid, XtPointer unused, XtPointer unused_too)
#endif
{
EditorPtr ep;

    ep = getEditor(wid);

    if (ep->enteredText)
	XtFree(ep->enteredText);
    ep->enteredText = XmTextFieldGetString(wid);
    if (strlen(ep->enteredText) == 0)
    {
	XtFree(ep->enteredText);
	ep->enteredText = NULL;
    }

    XtRemoveGrab(wid);
    XtUnmanageChild(wid);
    textPrimIPF(ep);
}

#ifdef _NO_PROTO
void textPrimBPF(ep, e)
EditorPtr ep; XButtonEvent *e;
#else
void textPrimBPF(EditorPtr ep, XButtonEvent *e)
#endif
{
Pixmap pix;
static GC bitmapgc = NULL;
int ascent, descent, dummy;
XCharStruct overall;

    if (e->button == 1)
	ep->fg = (int)curPixel;
    else if (e->button == 2)
    {
	getTextEntry(ep, e->x, e->y);
	return;
    }
    else
	ep->fg = 0;
    if (!ep->enteredText)
    {
	popupMessage(XPE_Warning, "No Text Entered!", NULL, NULL, NULL);
	return;
    }

    XQueryTextExtents(display, currentFont->fid, 
	ep->enteredText, strlen(ep->enteredText), &dummy, &ascent, &descent,
	&overall);
    ep->txth = ascent + descent;
    ep->txtw = overall.width;

    pix = XCreatePixmap(display, XtWindow(ep->workarea), ep->txtw, ep->txth, 1);

    if (!bitmapgc)
	bitmapgc = XCreateGC(display, pix, 0, NULL);

    /* clear out bitmap */
    XSetForeground(display, bitmapgc, 0);
    XFillRectangle(display, pix, bitmapgc, 0, 0, ep->txtw, ep->txth);

    /* draw text */
    XSetForeground(display, bitmapgc, 1);
    XSetFont(display, bitmapgc, currentFont->fid);
    XDrawString(display, pix, bitmapgc, 0, ascent, 
	ep->enteredText, strlen(ep->enteredText));

    if (ep->textimage)
	XDestroyImage(ep->textimage);

    /* get back an image */
    ep->textimage = XGetImage(display, pix, 0, 0, ep->txtw, ep->txth, 
	AllPlanes, ZPixmap);
    
    /* done with bitmap */
    XFreePixmap(display, pix);

    saveState(ep);
    ep->lastx = e->x;
    ep->lasty = e->y;
    blitBitmapImage(ep, ep->textimage, False, e->x, e->y, ep->txtw, ep->txth);
}

#ifdef _NO_PROTO
void textPrimBRF(ep, e)
EditorPtr ep; XButtonEvent *e;
#else
void textPrimBRF(EditorPtr ep, XButtonEvent *e)
#endif
{
    if (!ep->textimage || e->button == 2)
	return;
    blitBitmapImage(ep, ep->textimage, True, ep->lastx, ep->lasty, ep->txtw, ep->txth);
}

#ifdef _NO_PROTO
void textPrimBMF(ep, e)
EditorPtr ep; XMotionEvent *e;
#else
void textPrimBMF(EditorPtr ep, XMotionEvent *e)
#endif
{
    if (!ep->textimage || e->state & Button2Mask)
	return;
    if (e->x == ep->lastx && e->y == ep->lasty)
	return;
    unblitBitmapImage(ep, ep->textimage, ep->lastx, ep->lasty, ep->txtw, ep->txth);

    ep->lastx = e->x;
    ep->lasty = e->y;
    blitBitmapImage(ep, ep->textimage, False, e->x, e->y, ep->txtw, ep->txth);
}

/********************************************************************
** copyPrim class functions
********************************************************************/
#ifdef _NO_PROTO
void rubberBox(ep, x1, y1, x2, y2, drawit)
EditorPtr ep; int x1; int y1; int x2; int y2; int drawit;
#else
void rubberBox(EditorPtr ep, int x1, int y1, int x2, int y2, int drawit)
#endif
{
int n;
int x, y, w, h;
    if (x2 < x1)
    {
	n = x1; x1 = x2; x2 = n;
    }
    if (y2 < y1)
    {
	n = y1; y1 = y2; y2 = n;
    }
    if (x1 < 0)
	x1 = 0;
    if (y1 < 0)
	y1 = 0;
    if (x2 >= ep->pw)
	x2 = ep->pw - 1;
    if (y2 >= ep->ph)
	y2 = ep->ph - 1;

    x = x1 * (ep->scale + 1);
    y = y1 * (ep->scale + 1);
    w = (x2 - x1 + 1) * (ep->scale + 1);
    h = (y2 - y1 + 1) * (ep->scale + 1);

    if (drawit)
    {
	XSetForeground(display, ep->gc, ep->workfg);
	XDrawRectangle(display, XtWindow(ep->workarea), ep->gc, x, y, w, h);
	XDrawRectangle(display, XtWindow(ep->pixarea), ep->gc,
	    x1, y1, (x2 - x1), (y2 - y1));
    }
    else
    {
	XSetForeground(display, ep->gc, ep->workbg);
	XDrawRectangle(display, XtWindow(ep->workarea), ep->gc, x, y, w, h);
	XCopyArea(display, ep->pixmap, XtWindow(ep->pixarea), ep->gc, 
		x1, y1, (x2 - x1 + 1), (y2 - y1 + 1), x1, y1);
    }
}

/* ARGSUSED */
#ifdef _NO_PROTO
void handleClipboard(wid, dat, epPtr, reason)
Widget wid; int *dat; EditorPtr *epPtr; int *reason;
#else
void handleClipboard(Widget wid, int *dat, EditorPtr *epPtr, int *reason)
#endif
{
EditorPtr ep = *epPtr;
    if (*reason == XmCR_CLIPBOARD_DATA_REQUEST)
    {
	printf("what are we doing here?\n");
	return;
    }
    else	/* delete data */
    {
	ep->clipID = 0;
    }
}

#ifdef _NO_PROTO
void startClipboardCopy(ep, time)
EditorPtr ep; Time time;
#else
void startClipboardCopy(EditorPtr ep, Time time)
#endif
{
XmString xs;
int status;
    if (ep->clipID)
    {
	if (XmClipboardUndoCopy(display, XtWindow(ep->workarea))
		== ClipboardLocked)
	{
	    popupMessage(XPE_Error, 
		"Clipboard Locked!\nCannot Undo Old Copy",
		NULL, NULL, NULL);
	    return;
	}
	ep->clipID = 0;
    }
    xs = XS("XPE");
    status = XmClipboardStartCopy(display, XtWindow(ep->workarea),
	xs, time, ep->workarea, (VoidProc)handleClipboard, &(ep->clipID));
    XmStringFree(xs);
    if (status == ClipboardLocked)
    {
	popupMessage(XPE_Error, "Clipboard Locked!\nCannot Start Copy",
		NULL, NULL, NULL);
	return;
    }
}

#ifdef _NO_PROTO
int endClipboardCopy(ep, clearArea)
EditorPtr ep; int clearArea;
#else
int endClipboardCopy(EditorPtr ep, int clearArea)
#endif
{
int status;
int x1, y1, x2, y2, w, h;
    if (ep->lastx < ep->basex)
	{ x1 = ep->lastx; x2 = ep->basex; }
    else
	{ x1 = ep->basex; x2 = ep->lastx; }
    if (ep->lasty < ep->basey)
	{ y1 = ep->lasty; y2 = ep->basey; }
    else
	{ y1 = ep->basey; y2 = ep->lasty; }

    if (x1 < 0) x1 = 0;
    if (y1 < 0) y1 = 0;
    if (x2 >= ep->pw) x2 = ep->pw;
    if (y2 >= ep->ph) y2 = ep->ph;

    w = x2 - x1 + 1;
    h = y2 - y1 + 1;

    ep->clipboardPixmap = XCreatePixmap(display, XtWindow(ep->workarea), w, h, 
	    DefaultDepth(display, screen));
    XCopyArea(display, ep->pixmap, ep->clipboardPixmap, ep->gc, 
	x1, y1, w, h, 0, 0);

    status = XmClipboardCopy(display, XtWindow(ep->workarea), ep->clipID, 
	"PIXMAP", (char *)&(ep->clipboardPixmap), sizeof(Pixmap), 
	(int)ep, &(ep->dataID));
    if (status != ClipboardSuccess)
    {
	XmClipboardCancelCopy(display, XtWindow(ep->workarea), ep->clipID);
	popupMessage(XPE_Error, "Clipboard Copy Failed",
		NULL, NULL, NULL);
	ep->clipID = 0;
	XFreePixmap(display, ep->clipboardPixmap);
	ep->clipboardPixmap = 0;
	return (!Success);
    }
    if (XmClipboardEndCopy(display, XtWindow(ep->workarea), ep->clipID) 
	!= ClipboardSuccess)
    {
	XmClipboardCancelCopy(display, XtWindow(ep->workarea), ep->clipID);
	popupMessage(XPE_Error, "Clipboard End Copy Failed!",
		NULL, NULL, NULL);
	ep->clipID = 0;
	XFreePixmap(display, ep->clipboardPixmap);
	ep->clipboardPixmap = 0;
	return (!Success);
    }
    /* cut leaves area blank */
    if (clearArea)
    {
	XSetForeground(display, ep->gc, 0);
	XFillRectangle(display, ep->pixmap, ep->gc, x1, y1, w, h);
	XGetSubImage(display, ep->pixmap, 0, 0, ep->pw, ep->ph, 
	    AllPlanes, ZPixmap, ep->image, 0, 0);
	redrawWidget(ep->workarea);
    }

    return(Success);
}

#ifdef _NO_PROTO
void copyPrimBPF(ep, e)
EditorPtr ep; XButtonEvent *e;
#else
void copyPrimBPF(EditorPtr ep, XButtonEvent *e)
#endif
{
    startClipboardCopy(ep, e->time);
    saveState(ep);
    ep->lastx = ep->basex = e->x;
    ep->lasty = ep->basey = e->y;
    rubberBox(ep, e->x, e->y, e->x, e->y, True);
}

/* ARGSUSED */
#ifdef _NO_PROTO
void copyPrimBRF(ep, e)
EditorPtr ep; XButtonEvent *e;
#else
void copyPrimBRF(EditorPtr ep, XButtonEvent *e)
#endif
{
    if (endClipboardCopy(ep, False) != Success)
    {
	XFreePixmap(display, ep->clipboardPixmap);
	ep->clipboardPixmap = 0;
    }
    rubberBox(ep, ep->basex, ep->basey, ep->lastx, ep->lasty, False);
}

#ifdef _NO_PROTO
void copyPrimBMF(ep, e)
EditorPtr ep; XMotionEvent *e;
#else
void copyPrimBMF(EditorPtr ep, XMotionEvent *e)
#endif
{
    if (e->x == ep->lastx && e->y == ep->lasty)
	return;
    rubberBox(ep, ep->basex, ep->basey, ep->lastx, ep->lasty, False);
    ep->lastx = e->x;
    ep->lasty = e->y;
    rubberBox(ep, ep->basex, ep->basey, e->x, e->y, True);
}

/********************************************************************
** movePrim class functions
********************************************************************/
#ifdef _NO_PROTO
void movePrimBPF(ep, e)
EditorPtr ep; XButtonEvent *e;
#else
void movePrimBPF(EditorPtr ep, XButtonEvent *e)
#endif
{
    startClipboardCopy(ep, e->time);
    saveState(ep);
    ep->lastx = ep->basex = e->x;
    ep->lasty = ep->basey = e->y;
    rubberBox(ep, e->x, e->y, e->x, e->y, True);
}

/* ARGSUSED */
#ifdef _NO_PROTO
void movePrimBRF(ep, e)
EditorPtr ep; XButtonEvent *e;
#else
void movePrimBRF(EditorPtr ep, XButtonEvent *e)
#endif
{
    if (endClipboardCopy(ep, True) != Success)
    {
	XFreePixmap(display, ep->clipboardPixmap);
	ep->clipboardPixmap = 0;
    }
    rubberBox(ep, ep->basex, ep->basey, ep->lastx, ep->lasty, False);
}

#ifdef _NO_PROTO
void movePrimBMF(ep, e)
EditorPtr ep; XMotionEvent *e;
#else
void movePrimBMF(EditorPtr ep, XMotionEvent *e)
#endif
{
    if (e->x == ep->lastx && e->y == ep->lasty)
	return;
    rubberBox(ep, ep->basex, ep->basey, ep->lastx, ep->lasty, False);
    ep->lastx = e->x;
    ep->lasty = e->y;
    rubberBox(ep, ep->basex, ep->basey, e->x, e->y, True);
}

/********************************************************************
** pastePrim class functions
********************************************************************/
#ifdef _NO_PROTO
void pastePrimBPF(ep, e)
EditorPtr ep; XButtonEvent *e;
#else
void pastePrimBPF(EditorPtr ep, XButtonEvent *e)
#endif
{
Pixmap pix;
unsigned long num_bytes;
int id, status;
int dummy;
unsigned int uidummy, w, h;
Window root;

    ep->pasteImage = 0;
    status = XmClipboardRetrieve(display, XtWindow(ep->workarea), "PIXMAP",
	(char *)&pix, sizeof(Pixmap), &num_bytes, &id);
    if (status != ClipboardSuccess)
    {
	if (status == ClipboardNoData)
	    popupMessage(XPE_Warning, "No PIXMAP On Clipboard!",
		    NULL, NULL, NULL);
	else if (status == ClipboardLocked)
	    popupMessage(XPE_Error, "Clipboard Locked - Retrieve Failed!",
		    NULL, NULL, NULL);
	else
	    popupMessage(XPE_Error, "Clipboard Retrieve Failed!",
		    NULL, NULL, NULL);
	return;
    }
    saveState(ep);
    XGetGeometry(display, pix, &root, &dummy, &dummy,
	&w, &h, &uidummy, &uidummy);
    ep->pasteImage = XGetImage(display, pix, 0, 0, w, h,
		    AllPlanes, ZPixmap);

    if (e->button == 3)
	ep->opaquePaste = True;
    else
	ep->opaquePaste = False;

    ep->lastx = e->x;
    ep->lasty = e->y;
    blitImage(ep, ep->pasteImage, e->x, e->y, ep->opaquePaste, False);
}

/* ARGSUSED */
#ifdef _NO_PROTO
void pastePrimBRF(ep, e)
EditorPtr ep; XButtonEvent *e;
#else
void pastePrimBRF(EditorPtr ep, XButtonEvent *e)
#endif
{
    if (!ep->pasteImage)
	return;
    blitImage(ep, ep->pasteImage, ep->lastx, ep->lasty, ep->opaquePaste, True);
    XDestroyImage(ep->pasteImage);
}

#ifdef _NO_PROTO
void pastePrimBMF(ep, e)
EditorPtr ep; XMotionEvent *e;
#else
void pastePrimBMF(EditorPtr ep, XMotionEvent *e)
#endif
{
    if (!ep->pasteImage)
	return;
    if (e->x == ep->lastx && e->y == ep->lasty)
	return;
    unblitImage(ep, ep->pasteImage, ep->lastx, ep->lasty, ep->opaquePaste);
    ep->lastx = e->x;
    ep->lasty = e->y;
    blitImage(ep, ep->pasteImage, e->x, e->y, ep->opaquePaste, False);
}

/********************************************************************
** circlePrim class functions
********************************************************************/
#ifdef _NO_PROTO
void drawCircle(ep, x, y, state, filled, xret, yret, dxret, dyret)
EditorPtr ep; int x; int y; unsigned int state; int filled;
    int *xret; int *yret; int *dxret; int *dyret;
#else
void drawCircle(EditorPtr ep, int x, int y, unsigned int state, int filled,
    int *xret, int *yret, int *dxret, int *dyret)
#endif
{
    int dx, dy;

    dx = ep->basex - x; if (dx < 0) dx = -dx;
    dy = ep->basey - y; if (dy < 0) dy = -dy;
    if (state & ShiftMask)
    {
	if (dx > dy)
	    dy = dx;
	else
	    dx = dy;
    }
    x = ep->basex - dx;
    y = ep->basey - dy;
    dx = dx * 2;
    dy = dy * 2;
    /* we will always draw a circle even if we are filling */
    /* because it makes the filled arc look better */
    XDrawArc(display, ep->workpixmap, ep->workgc,
	    x, y, dx, dy, 0, 64 * 360);
    if (filled)
	XFillArc(display, ep->workpixmap, ep->workgc,
		x, y, dx, dy, 0, 64 * 360);
    /* fill in return parameters */
    dx++;
    dy++;
    if (x < 0) x = 0;
    if (y < 0) y = 0;
    if (x + dx > ep->pw) dx = ep->pw - x;
    if (y + dy > ep->ph) dy = ep->ph - y;
    *xret = x;
    *yret = y;
    *dxret = dx;
    *dyret = dy;
}

#ifdef _NO_PROTO
void circlePrimBPF(ep, e)
EditorPtr ep; XButtonEvent *e;
#else
void circlePrimBPF(EditorPtr ep, XButtonEvent *e)
#endif
{
    int cx, cy, cdx, cdy;

    saveState(ep);
    if (e->button == 1)
	ep->fg = (int)curPixel;
    else if (e->button == 2)
    {
	ep->fg = (int)XGetPixel(ep->image, e->x, e->y);
	setCurrentPixel(ep->fg);
    }
    else
	ep->fg = 0;

    ep->lastx = ep->basex = e->x;
    ep->lasty = ep->basey = e->y;
    clearWorkPixmap(ep);
    XSetForeground(display, ep->workgc, 1);
    drawCircle(ep, e->x, e->y, e->state, False, &cx, &cy, &cdx, &cdy);
    blitWorkPixmap(ep, False, False, cx, cy, cdx, cdy);
}

/* ARGSUSED */
#ifdef _NO_PROTO
void circlePrimBRF(ep, e)
EditorPtr ep; XButtonEvent *e;
#else
void circlePrimBRF(EditorPtr ep, XButtonEvent *e)
#endif
{
    blitWorkPixmap(ep, True, False, 0, 0, ep->pw, ep->ph);
}

#ifdef _NO_PROTO
void circlePrimBMF(ep, e)
EditorPtr ep; XMotionEvent *e;
#else
void circlePrimBMF(EditorPtr ep, XMotionEvent *e)
#endif
{
    int cx, cy, cdx, cdy;
    if (e->x == ep->lastx && e->y == ep->lasty)
	return;
    unblitWorkPixmap(ep, 0, 0, ep->pw, ep->ph);

    ep->lastx = e->x;
    ep->lasty = e->y;
    clearWorkPixmap(ep);
    drawCircle(ep, e->x, e->y, e->state, False, &cx, &cy, &cdx, &cdy);
    blitWorkPixmap(ep, False, False, cx, cy, cdx, cdy);
}


/********************************************************************
** discPrim class functions
********************************************************************/
#ifdef _NO_PROTO
void discPrimBPF(ep, e)
EditorPtr ep; XButtonEvent *e;
#else
void discPrimBPF(EditorPtr ep, XButtonEvent *e)
#endif
{
    int cx, cy, cdx, cdy;
    saveState(ep);
    if (e->button == 1)
	ep->fg = (int)curPixel;
    else if (e->button == 2)
    {
	ep->fg = (int)XGetPixel(ep->image, e->x, e->y);
	setCurrentPixel(ep->fg);
    }
    else
	ep->fg = 0;

    ep->lastx = ep->basex = e->x;
    ep->lasty = ep->basey = e->y;
    clearWorkPixmap(ep);
    XSetForeground(display, ep->workgc, 1);
    drawCircle(ep, e->x, e->y, e->state, True, &cx, &cy, &cdx, &cdy);
    blitWorkPixmap(ep, False, False, cx, cy, cdx, cdy);
}

/* ARGSUSED */
#ifdef _NO_PROTO
void discPrimBRF(ep, e)
EditorPtr ep; XButtonEvent *e;
#else
void discPrimBRF(EditorPtr ep, XButtonEvent *e)
#endif
{
    blitWorkPixmap(ep, True, False, 0, 0, ep->pw, ep->ph);
}

#ifdef _NO_PROTO
void discPrimBMF(ep, e)
EditorPtr ep; XMotionEvent *e;
#else
void discPrimBMF(EditorPtr ep, XMotionEvent *e)
#endif
{
    int cx, cy, cdx, cdy;
    if (e->x == ep->lastx && e->y == ep->lasty)
	return;
    unblitWorkPixmap(ep, 0, 0, ep->pw, ep->ph);

    ep->lastx = e->x;
    ep->lasty = e->y;
    clearWorkPixmap(ep);
    drawCircle(ep, e->x, e->y, e->state, True, &cx, &cy, &cdx, &cdy);
    blitWorkPixmap(ep, False, False, cx, cy, cdx, cdy);
}


/*
** primitive selection routine
*/

typedef struct 
{
    String		funcName;
    int			cursorIndex;
    Cursor		cursor;
    ButtonFunction	initFunc;
    ButtonFunction	pressFunc;
    ButtonFunction	releaseFunc;
    ButtonFunction	motionFunc;
} PrimitiveTable, *PrimitiveTablePtr;

static PrimitiveTable primTable[] = 
{
    { "primDraw",	XC_pencil, NULL,
		NULL, drawPrimBPF, drawPrimBRF, drawPrimBMF},
    { "primLine",	XC_cross, NULL,
		NULL, linePrimBPF, linePrimBRF, linePrimBMF},
    { "primRect",	XC_lr_angle, NULL,
		NULL, rectPrimBPF, rectPrimBRF, rectPrimBMF},
    { "primFillRect",	XC_lr_angle, NULL,
		NULL, fillRectPrimBPF,fillRectPrimBRF,fillRectPrimBMF},
    { "primCircle",	XC_circle, NULL,
		NULL, circlePrimBPF, circlePrimBRF, circlePrimBMF},
    { "primFillCircle", XC_circle, NULL,
		NULL, discPrimBPF, discPrimBRF, discPrimBMF},
    { "primFlood",	XC_box_spiral, NULL,
		floodPrimIPF, floodPrimBPF, floodPrimBRF, floodPrimBMF},
    { "primBitmap",	XC_gumby, NULL,
		bitmapPrimIPF, bitmapPrimBPF, bitmapPrimBRF, bitmapPrimBMF},
    { "primText",	XC_xterm, NULL,
		textPrimIPF, textPrimBPF, textPrimBRF, textPrimBMF},
    { "primCopy",	XC_bottom_right_corner, NULL,
		NULL, copyPrimBPF, copyPrimBRF, copyPrimBMF},
    { "primMove",	XC_pirate, NULL,
		NULL, movePrimBPF, movePrimBRF, movePrimBMF},
    { "primPaste",	XC_hand2, NULL,
		NULL, pastePrimBPF, pastePrimBRF, pastePrimBMF},
    { NULL, NULL, NULL, NULL, NULL}
};

/* ARGSUSED */
#ifdef _NO_PROTO
void selectPrimCB(wid, unused, cb)
Widget wid; XtPointer unused; XmToggleButtonCallbackStruct *cb;
#else
void selectPrimCB(Widget wid, XtPointer unused, 
	XmToggleButtonCallbackStruct *cb)
#endif
{
EditorPtr ep;
PrimitiveTablePtr ptp = primTable;
String name = XtName(wid);

    if (cb && !cb->set)
	return;

    ep = getEditor(wid);
    while (ptp->funcName)
    {
	if (strcmp(name, ptp->funcName) == 0)
	{
	    sayMessage(ep, "");
	    if (ptp->initFunc)
		(*(ptp->initFunc))(ep);
	    if (!ptp->cursor)
		ptp->cursor = XCreateFontCursor(display, ptp->cursorIndex);
	    ep->buttonPressFunction = ptp->pressFunc;
	    ep->buttonReleaseFunction = ptp->releaseFunc;
	    ep->buttonMotionFunction = ptp->motionFunc;
	    XDefineCursor(display, XtWindow(ep->workarea), ptp->cursor);
	    XDefineCursor(display, XtWindow(ep->pixarea), ptp->cursor);
	    break;
	}
	ptp++;
    }
}
